#    Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License").
#    You may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

import json
import logging
import os
import time
from typing import Any, Dict, List, Optional, Tuple, cast

import boto3
from cfn_tools import load_yaml

import seedfarmer.commands._seedkit_commands as sk_commands
import seedfarmer.errors
import seedfarmer.services._cfn as cfn
import seedfarmer.services._iam as iam
import seedfarmer.services._s3 as s3
from seedfarmer import config
from seedfarmer.mgmt.bundle_support import BUNDLE_PREFIX
from seedfarmer.mgmt.module_info import get_module_stack_names
from seedfarmer.models.manifests import DeploymentManifest, ModuleParameter
from seedfarmer.services import get_sts_identity_info
from seedfarmer.services.session_manager import SessionManager
from seedfarmer.types.parameter_types import EnvVar, EnvVarType
from seedfarmer.utils import generate_hash, upper_snake_case

_logger: logging.Logger = logging.getLogger(__name__)


class StackInfo(object):
    _PROJECT_MANAGED_POLICY_CFN_NAME: Optional[str] = None
    _SEEDFARMER_BUCKET_CFN_NAME: Optional[str] = None
    _region: Optional[str] = None
    _account_id: Optional[str] = None

    def __init__(self, account_id: Optional[str] = None, region: Optional[str] = None) -> None:
        self._account_id = account_id
        self._region = region

    @property
    def PROJECT_MANAGED_POLICY_CFN_NAME(self) -> str:
        if self._PROJECT_MANAGED_POLICY_CFN_NAME is None:
            self._PROJECT_MANAGED_POLICY_CFN_NAME = f"{config.PROJECT.lower()}-managed-policy"
        return self._PROJECT_MANAGED_POLICY_CFN_NAME

    @property
    def SEEDFARMER_BUCKET_STACK_NAME(self) -> str:
        return f"seedfarmer-{config.PROJECT.lower()}-artifacts"


info = StackInfo()


def _get_project_managed_policy_arn(session: Optional[boto3.Session]) -> str:
    def _check_stack_status() -> Tuple[bool, Dict[str, str]]:
        return cfn.does_stack_exist(stack_name=info.PROJECT_MANAGED_POLICY_CFN_NAME, session=session)

    retries = 3
    while retries > 0:
        project_managed_policy_stack_exists, stack_outputs = _check_stack_status()
        if project_managed_policy_stack_exists:
            if stack_outputs.get("StackStatus") and "_IN_PROGRESS" in stack_outputs.get("StackStatus"):  # type: ignore
                _logger.info("The managed policy stack is not complete, waiting 30 seconds")
                time.sleep(30)
                retries -= 1
            else:
                _logger.debug("project_managed_policy_output is : %s", stack_outputs)
                project_managed_policy_arn = stack_outputs.get("ProjectPolicyARN", None)
                retries = -1
        else:
            _logger.debug("project_managed_policy_output does not exist")
            retries = -1

    if not project_managed_policy_arn:
        raise seedfarmer.errors.InvalidConfigurationError(
            "Project Managed Stack is missing the export `ProjectPolicyARN`"
        )
    else:
        return project_managed_policy_arn


def _get_seedkit_resources_policy_arn(session: Optional[boto3.Session]) -> Optional[str]:
    seedkit_stack_exists, seedkit_stack_name, stack_outputs = sk_commands.seedkit_deployed(
        seedkit_name=config.PROJECT, session=session
    )
    if seedkit_stack_exists:
        return cast(str, stack_outputs.get("SeedkitResourcesPolicyArn"))
    return None


def _get_docker_secret_inline_policy(docker_credentials_secret: str, session: Optional[boto3.Session]) -> str:
    account_id, region, partition = get_sts_identity_info(session=session)
    return json.dumps(
        {
            "Version": "2012-10-17",
            "Statement": [
                {
                    "Effect": "Allow",
                    "Action": [
                        "secretsmanager:GetSecretValue",
                        "secretsmanager:DescribeSecret",
                        "secretsmanager:ListSecretVersionIds",
                    ],
                    "Resource": (
                        f"arn:{partition}:secretsmanager:{region}:{account_id}:secret:{docker_credentials_secret}*"
                    ),
                },
                {"Effect": "Allow", "Action": ["secretsmanager:ListSecrets"], "Resource": "*"},
            ],
        }
    )


def create_module_deployment_role(
    role_name: str,
    deployment_name: str,
    group_name: Optional[str] = None,
    module_name: Optional[str] = None,
    docker_credentials_secret: Optional[str] = None,
    permissions_boundary_arn: Optional[str] = None,
    session: Optional[boto3.Session] = None,
    role_prefix: Optional[str] = None,
) -> None:
    iam.create_check_iam_role(
        project_name=config.PROJECT,
        deployment_name=deployment_name,
        group_name=group_name,
        module_name=module_name,
        trust_policy={
            "Version": "2012-10-17",
            "Statement": [
                {"Effect": "Allow", "Principal": {"Service": "codebuild.amazonaws.com"}, "Action": "sts:AssumeRole"}
            ],
        },
        role_name=role_name,
        permissions_boundary_arn=permissions_boundary_arn,
        session=session,
        role_prefix=role_prefix,
    )

    policies = []
    seedkit_resources_policy_arn = _get_seedkit_resources_policy_arn(session=session)
    if seedkit_resources_policy_arn:
        policies.append(seedkit_resources_policy_arn)

    project_managed_policy_arn = _get_project_managed_policy_arn(session=session)
    policies.append(project_managed_policy_arn)

    _logger.debug(
        f"seedkit_resources_policy {seedkit_resources_policy_arn}  project_managed_policy {project_managed_policy_arn}"
    )

    policies_attached = iam.attach_policy_to_role(role_name, policies, session=session)
    if policies.sort() == policies_attached.sort():
        _logger.info("Delaying deployment to allow %s IAM Role and Policies to take effect", role_name)
        time.sleep(12)

    if docker_credentials_secret:
        policy_body = _get_docker_secret_inline_policy(
            docker_credentials_secret=docker_credentials_secret, session=session
        )
        iam.attach_inline_policy(
            role_name=role_name, policy_body=policy_body, policy_name=docker_credentials_secret, session=session
        )


def destroy_module_deployment_role(
    role_name: str,
    docker_credentials_secret: Optional[str] = None,
    session: Optional[boto3.Session] = None,
) -> None:
    policies = []
    seedkit_resources_policy_arn = _get_seedkit_resources_policy_arn(session=session)
    if seedkit_resources_policy_arn:
        policies.append(seedkit_resources_policy_arn)

    # Extract Project Managed policy name
    project_managed_policy_stack_exists, stack_outputs = cfn.does_stack_exist(
        stack_name=info.PROJECT_MANAGED_POLICY_CFN_NAME, session=session
    )
    if project_managed_policy_stack_exists:
        project_managed_policy_arn = stack_outputs.get("ProjectPolicyARN")
        policies.append(str(project_managed_policy_arn))

    _logger.debug(
        f"seedkit_resources_policy {seedkit_resources_policy_arn}  project_managed_policy {project_managed_policy_arn}"
    )

    for policy_arn in policies:
        iam.detach_policy_from_role(role_name=role_name, policy_arn=policy_arn, session=session)

    # Detach Docker secret policy
    if docker_credentials_secret:
        iam.detach_inline_policy_from_role(role_name=role_name, policy_name=docker_credentials_secret, session=session)

    iam.delete_role(role_name=role_name, session=session)


def deploy_bucket_storage_stack(
    account_id: str,
    region: str,
    **kwargs: Any,
) -> str:
    """
    deploy_bucket_storage_stack
        This function deploys the account/region bucket for storing artifacts

    Parameters
    ----------
    account_id: str
        The Account Id where the module is deployed
    region: str
        The region where the module is deployed

    Returns
    -------
    str
        the bucket name
    """

    session = SessionManager().get_or_create().get_deployment_session(account_id=account_id, region_name=region)
    info = StackInfo(account_id=account_id, region=region)
    bucket_stack_name = info.SEEDFARMER_BUCKET_STACK_NAME

    stack_exists, output = cfn.does_stack_exist(stack_name=bucket_stack_name, session=session)

    if not stack_exists:
        _logger.info("Deploying the bucket storage stack %s", bucket_stack_name)
        hash = generate_hash(string=f"{account_id}-{region}", length=6)
        bucket_name = f"seedfarmer-{config.PROJECT.lower()}-{region}-{account_id}-{hash}-no-delete"
        if len(bucket_name) > 63:
            limit = len(f"seedfarmer--{config.PROJECT.lower()}-no-delete")
            sm_hash = generate_hash(string=f"{config.PROJECT.lower()}-{region}-{account_id}", length=63 - limit)
            bucket_name = f"seedfarmer-{config.PROJECT.lower()}-{sm_hash}-no-delete"
        cfn.deploy_template(
            stack_name=bucket_stack_name,
            filename=config.BUCKET_STORAGE_PATH,
            seedkit_tag=config.PROJECT.lower(),
            parameters={"ProjectName": config.PROJECT.lower(), "BucketName": bucket_name[:63]},
            session=session,
        )
        stack_exists, output = cfn.does_stack_exist(stack_name=bucket_stack_name, session=session)
    return str(output.get("Bucket"))


def deploy_managed_policy_stack(
    deployment_manifest: DeploymentManifest,
    account_id: str,
    region: str,
    update_project_policy: Optional[bool] = False,
    **kwargs: Any,
) -> None:
    """
    deploy_managed_policy_stack
        This function deploys the deployment-specific policy

    Parameters
    ----------
    deployment_manifest : DeploymentManifest
        The DeploymentManifest object of the deploy
    account_id: str
        The Account Id where the module is deployed
    region: str
        The region where the module is deployed
    update_project_policy: bool
        Force update the project policy if already deployed
    """
    # Determine if managed policy stack already deployed
    session = SessionManager().get_or_create().get_deployment_session(account_id=account_id, region_name=region)
    project_managed_policy_stack_exists, _ = cfn.does_stack_exist(
        stack_name=info.PROJECT_MANAGED_POLICY_CFN_NAME, session=session
    )
    if not project_managed_policy_stack_exists or update_project_policy:
        project_managed_policy_template = config.PROJECT_POLICY_PATH
        _logger.info("Resolved the ProjectPolicyPath %s", project_managed_policy_template)
        if not os.path.exists(project_managed_policy_template):
            raise seedfarmer.errors.InvalidPathError(
                f"Unable to find the Project Managed Policy Template: {project_managed_policy_template}"
            )
        _logger.info(
            "Deploying %s from the path %s", info.PROJECT_MANAGED_POLICY_CFN_NAME, project_managed_policy_template
        )
        cfn.deploy_template(
            stack_name=info.PROJECT_MANAGED_POLICY_CFN_NAME,
            filename=project_managed_policy_template,
            seedkit_tag=deployment_manifest.name,
            parameters={"ProjectName": config.PROJECT.lower(), "DeploymentName": str(deployment_manifest.name)},
            session=session,
        )


def destroy_bucket_storage_stack(
    account_id: str,
    region: str,
    **kwargs: Any,
) -> None:
    """
    destroy_bucket_storage_stack
        This function destroys the bucket stack for SeedFarmer

    Parameters
    ----------
    account_id: str
        The Account Id where the module is deployed
    region: str
        The region where deployed
    """

    # Determine if managed policy stack already deployed
    session = SessionManager().get_or_create().get_deployment_session(account_id=account_id, region_name=region)
    info = StackInfo(account_id=account_id, region=region)
    bucket_stack_name = info.SEEDFARMER_BUCKET_STACK_NAME
    bucket_stack_exists, outputs = cfn.does_stack_exist(stack_name=bucket_stack_name, session=session)
    if bucket_stack_exists and outputs is not None:
        bucket_name = str(outputs.get("Bucket"))
        bucket_empty = s3.is_bucket_empty(bucket=bucket_name, folder=BUNDLE_PREFIX, session=session)
        if bucket_empty:
            _logger.info("Destroying Stack %s in Account/Region: %s/%s", bucket_stack_name, account_id, region)
            import botocore.exceptions

            try:
                s3.delete_objects(bucket=bucket_name, session=session)
                cfn.destroy_stack(
                    stack_name=bucket_stack_name,
                    session=session,
                )
            except (botocore.exceptions.WaiterError, botocore.exceptions.ClientError):
                _logger.info(f"Failed to delete project stack {bucket_stack_name}, ignoring and moving on")
        else:
            _logger.info("Stack %s left, S3 bucket %s not empty", bucket_stack_name, bucket_name)
            return


def destroy_managed_policy_stack(account_id: str, region: str, **kwargs: Any) -> None:
    """
    destroy_managed_policy_stack
        This function destroys the deployment-specific policy.

    Parameters
    ----------
    account_id: str
        The Account Id where the module is deployed
    region: str
        The region where the module is deployed
    """
    # Determine if managed policy stack already deployed
    session = SessionManager().get_or_create().get_deployment_session(account_id=account_id, region_name=region)
    project_managed_policy_stack_exists, stack_outputs = cfn.does_stack_exist(
        stack_name=info.PROJECT_MANAGED_POLICY_CFN_NAME, session=session
    )
    _logger.debug("project_managed_policy_output is : %s", stack_outputs)
    has_roles_attached = False
    if project_managed_policy_stack_exists:
        project_managed_policy_arn = stack_outputs.get("ProjectPolicyARN")
        policy = iam.get_policy_info(policy_arn=str(project_managed_policy_arn), session=session)
        has_roles_attached = True if policy and policy["Policy"]["AttachmentCount"] > 0 else False

    if project_managed_policy_stack_exists and not has_roles_attached:
        _logger.info(
            "Destroying Stack %s in Account/Region: %s/%s", info.PROJECT_MANAGED_POLICY_CFN_NAME, account_id, region
        )
        import botocore.exceptions

        try:
            cfn.destroy_stack(
                stack_name=info.PROJECT_MANAGED_POLICY_CFN_NAME,
                session=session,
            )
        except (botocore.exceptions.WaiterError, botocore.exceptions.ClientError):
            _logger.info(
                f"Failed to delete project stack {info.PROJECT_MANAGED_POLICY_CFN_NAME}, ignoring and moving on"
            )
    else:
        _logger.info(
            "Stack %s in Account/Region: %s/%s is either not deployed or has roles attached",
            info.PROJECT_MANAGED_POLICY_CFN_NAME,
            account_id,
            region,
        )


def destroy_module_stack(
    deployment_name: str,
    group_name: str,
    module_name: str,
    account_id: str,
    region: str,
    docker_credentials_secret: Optional[str] = None,
) -> None:
    """
    destroy_module_stack
        This function destroys the module-specific stack created to support deployment

    Parameters
    ----------
    deployment : str
        The name of the deployment
    group : str
        The name of the group
    module_name : str
        The name of the module
    account_id: str
        The Account Id where the module is deployed
    region: str
        The region where the module is deployed
    """
    session = SessionManager().get_or_create().get_deployment_session(account_id=account_id, region_name=region)
    module_stack_name, module_role_name = get_module_stack_names(
        deployment_name, group_name, module_name, session=session
    )

    cfn.destroy_stack(module_stack_name, session=session)

    destroy_module_deployment_role(
        role_name=module_role_name,
        docker_credentials_secret=docker_credentials_secret,
        session=session,
    )


def deploy_module_stack(
    module_stack_path: str,
    deployment_name: str,
    group_name: str,
    module_name: str,
    account_id: str,
    region: str,
    parameters: List[ModuleParameter],
    docker_credentials_secret: Optional[str] = None,
    permissions_boundary_arn: Optional[str] = None,
    role_prefix: Optional[str] = None,
) -> Tuple[str, str]:
    """
    deploy_module_stack
        This function deploys the module stack (modulestack.yaml) to support the module
        for deployment

    Parameters
    ----------
    module_stack_path : str
        The path to the modulestack.yaml
    deployment : str
        The name of the deployment
    group : str
        The name of the group
    module : str
        The name of the module
    parameters : List[ModuleParameter]
        A ModuleParameter object with any necessary parameters from other deployed modules
    account_id: str
        The Account Id where the module is deployed
    region: str
        The region where the module is deployed
    docker_credentials_secret: str
        OPTIONAL parameter with name of SecrestManager of docker credentials
    permissions_boundary_arn: str
        OPTIONAL parameter with Name of PermissionBoundary ManagedPolicy
    """

    _logger.debug(module_stack_path)

    session = (
        SessionManager()
        .get_or_create(role_prefix=role_prefix)
        .get_deployment_session(account_id=account_id, region_name=region)
    )
    module_stack_name, module_role_name = get_module_stack_names(
        deployment_name, group_name, module_name, session=session
    )

    create_module_deployment_role(
        role_name=module_role_name,
        deployment_name=deployment_name,
        group_name=group_name,
        module_name=module_name,
        docker_credentials_secret=docker_credentials_secret,
        permissions_boundary_arn=permissions_boundary_arn,
        session=session,
        role_prefix=role_prefix,
    )

    _logger.debug("module_role_name %s", module_role_name)

    with open(module_stack_path, "r") as file:
        template = load_yaml(file)
        template_parameters = template.get("Parameters", {})

    stack_parameters = {}
    group_module_name = f"{group_name}-{module_name}"
    upper_snake_case_parameters = {
        **{p.upper_snake_case: p.value for p in parameters},
        **{
            "PROJECT_NAME": config.PROJECT,
            "DEPLOYMENT_NAME": deployment_name,
            "MODULE_NAME": group_module_name,
            "ROLE_NAME": module_role_name,
        },
    }
    for k in template_parameters.keys():
        upper_snake_case_key = upper_snake_case(k)
        if upper_snake_case_key in upper_snake_case_parameters:
            value = upper_snake_case_parameters[upper_snake_case_key]
            if isinstance(value, str):
                stack_parameters[k] = value
            elif isinstance(value, list):
                stack_parameters[k] = ",".join(value)
            elif isinstance(value, EnvVar):
                if (
                    value.type in [EnvVarType.PARAMETER_STORE.value, EnvVarType.SECRETS_MANAGER.value]
                    and ":" in value.value
                ):
                    stack_parameters[k] = value.value.split(":")[0]
                else:
                    stack_parameters[k] = value.value
            else:
                json.dumps(value)
    _logger.debug("stack_parameters: %s", stack_parameters)

    # Create/Update Module IAM Policy
    _logger.info("Deploying Module Stack for %s", group_module_name)
    cfn.deploy_template(
        stack_name=module_stack_name,
        filename=module_stack_path,
        seedkit_tag=module_stack_name,
        parameters=stack_parameters,
        session=session,
    )

    return module_stack_name, module_role_name


def get_module_stack_info(
    deployment_name: str,
    group_name: str,
    module_name: str,
    account_id: str,
    region: str,
) -> Tuple[str, str, bool]:
    """
    get_module_stack_info
        This function returns the name of the role and the name of the stack associated with the
        module deployment role

    Parameters
    ----------
    deployment_name : str
        Deployment Name
    group_name : str
        Group name
    module_name : str
        Module Name
    account_id : str
        The account id where deployed
    region : str
        The region where deployed

    Returns
    -------
    Tuple[str, str]
        A tuple with the  module_stack_name and  module_role_name
        [ module_stack_name, module_role_name ]
    """

    session = SessionManager().get_or_create().get_deployment_session(account_id=account_id, region_name=region)
    module_stack_name, module_role_name = get_module_stack_names(
        deployment_name, group_name, module_name, session=session
    )
    stack_exists, _ = cfn.does_stack_exist(stack_name=module_stack_name, session=session)
    return module_stack_name, module_role_name, stack_exists


def deploy_seedkit(
    account_id: str,
    region: str,
    vpc_id: Optional[str] = None,
    private_subnet_ids: Optional[List[str]] = None,
    security_group_ids: Optional[List[str]] = None,
    update_seedkit: Optional[bool] = False,
    role_prefix: Optional[str] = None,
    policy_prefix: Optional[str] = None,
    permissions_boundary_arn: Optional[str] = None,
    deploy_codeartifact: Optional[bool] = False,
    **kwargs: Any,
) -> Dict[str, Any]:
    """
    deploy_seedkit
        Deploy the SeedKit if not deployed

    Parameters
    ----------
    account_id: str
        The Account Id where the seedkit is deployed
    region: str
        The region where seedkit is deployed
    vpc_id: Optional[str]
        The VPC to associate seedkit with (codebuild)
    private_subnet_ids: Optional[List[str]]
        The Subnet IDs to associate seedkit with (codebuild)
    security_group_ids: Optional[List[str]]
        The Security Group IDs to associate seedkit with (codebuild)
    role_prefix: Optional[str]
        The IAM Path Prefix to use for seedkit role
    policy_prefix: Optional[str]
        The IAM Path Prefix to use for seedkit policy
    permissions_boundary_arn: Optional[str]
        The ARN of the permissions boundary to attach to seedkit role
    """
    session = SessionManager().get_or_create().get_deployment_session(account_id=account_id, region_name=region)
    stack_exists, _, stack_outputs = sk_commands.seedkit_deployed(seedkit_name=config.PROJECT, session=session)
    deploy_codeartifact = bool(stack_outputs.get("CodeArtifactRepository")) or bool(deploy_codeartifact)

    if stack_exists and not update_seedkit:
        _logger.debug("SeedKit exists and not updating for Account/Region: %s/%s", account_id, region)
    else:
        _logger.debug("Initializing / Updating SeedKit for Account/Region: %s/%s", account_id, region)

        seedkit_args = {
            "seedkit_name": config.PROJECT,
            "deploy_codeartifact": deploy_codeartifact,
            "session": session,
            "vpc_id": vpc_id,
            "subnet_ids": private_subnet_ids,
            "security_group_ids": security_group_ids,
        }

        if role_prefix:
            seedkit_args["role_prefix"] = role_prefix
        if policy_prefix:
            seedkit_args["policy_prefix"] = policy_prefix
        if permissions_boundary_arn:
            seedkit_args["permissions_boundary_arn"] = permissions_boundary_arn

        sk_commands.deploy_seedkit(**seedkit_args)  # type: ignore [arg-type]
        # Go get the outputs and return them
        _, _, stack_outputs = sk_commands.seedkit_deployed(seedkit_name=config.PROJECT, session=session)
    return dict(stack_outputs)


def destroy_seedkit(account_id: str, region: str) -> None:
    """
    destroy_seedkit
        Destroy the SeedKit if deployed

    Parameters
    ----------
    account_id: str
        The Account Id where the module is deployed
    region: str
        The region where the module is deployed
    """
    session = SessionManager().get_or_create().get_deployment_session(account_id=account_id, region_name=region)
    _logger.debug("Destroying SeedKit for Account/Region: %s/%s", account_id, region)
    sk_commands.destroy_seedkit(seedkit_name=config.PROJECT, session=session)


def force_manage_policy_attach(
    deployment_name: str,
    group_name: str,
    module_name: str,
    account_id: str,
    region: str,
    module_role_name: Optional[str] = None,
) -> None:
    session = SessionManager().get_or_create().get_deployment_session(account_id=account_id, region_name=region)
    if not module_role_name:
        module_stack_name, module_role_name = get_module_stack_names(
            deployment_name, group_name, module_name, session=session
        )

    project_managed_policy_arn = _get_project_managed_policy_arn(session=session)

    policies = [project_managed_policy_arn]
    iam.attach_policy_to_role(module_role_name, policies, session=session)



from typing import Any, Dict, List, Optional, cast

import seedfarmer.errors
from seedfarmer.models._base import CamelModel
from seedfarmer.models.manifests._deployment_manifest import DeploymentManifest
from seedfarmer.models.manifests._module_manifest import ModuleManifest, ModuleParameter


class ModuleDeployObject(CamelModel):
    deployment_manifest: DeploymentManifest
    group_name: Optional[str] = None
    module_name: Optional[str] = None
    module_manifest: Optional[ModuleManifest] = None
    parameters: Optional[List[ModuleParameter]] = None
    module_metadata: Optional[str] = None
    docker_credentials_secret: Optional[str] = None
    permissions_boundary_arn: Optional[str] = None
    module_role_name: Optional[str] = None
    module_role_arn: Optional[str] = None
    codebuild_image: Optional[str] = None
    runtime_overrides: Optional[Dict[str, str]] = None
    npm_mirror: Optional[str] = None
    npm_mirror_secret: Optional[str] = None
    pypi_mirror: Optional[str] = None
    pypi_mirror_secret: Optional[str] = None
    seedfarmer_bucket: Optional[str] = None

    def __init__(self, **kwargs: Any) -> None:
        super().__init__(**kwargs)

        if (
            (self.group_name and self.module_name and self.module_manifest)
            or ((self.group_name and not self.module_name) or (self.module_name and not self.group_name))
            or (not self.group_name and not self.module_name and not self.module_manifest)
        ):
            raise seedfarmer.errors.InvalidConfigurationError(
                "You must provide EITHER both 'group_name' and 'module_name', OR 'module_manifest'"
            )

        if self.group_name and self.module_name:
            _module = cast(ModuleManifest, self.deployment_manifest.get_module(self.group_name, self.module_name))

            pba = self.deployment_manifest.get_permission_boundary_arn(
                target_account=cast(str, _module.get_target_account_id()),
                target_region=cast(str, _module.target_region),
            )
            codebuild_image = self.deployment_manifest.get_region_codebuild_image(
                account_alias=_module.target_account, region=_module.target_region
            )
            region_runtime_overrides = self.deployment_manifest.get_region_runtime_overrides(
                account_alias=_module.target_account, region=_module.target_region
            )
            runtime_overrides = {
                **(region_runtime_overrides if region_runtime_overrides else {}),
                **(_module.runtime_overrides if _module.runtime_overrides else {}),
            }
            dcs = self.deployment_manifest.get_parameter_value(
                "dockerCredentialsSecret",
                account_alias=_module.target_account,
                region=_module.target_region,
            )
            npm_mirror = self.deployment_manifest.get_region_npm_mirror(
                account_alias=_module.target_account, region=_module.target_region
            )

            npm_mirror_secret = self.deployment_manifest.get_region_mirror_secret(
                account_alias=_module.target_account, region=_module.target_region, mirror_type="npm"
            )

            pypi_mirror = self.deployment_manifest.get_region_pypi_mirror(
                account_alias=_module.target_account, region=_module.target_region
            )

            pypi_mirror_secret = self.deployment_manifest.get_region_mirror_secret(
                account_alias=_module.target_account, region=_module.target_region, mirror_type="pypi"
            )

            sf_bucket = self.deployment_manifest.get_region_seedfarmer_bucket(
                account_alias=_module.target_account, region=_module.target_region
            )

            self.permissions_boundary_arn = pba
            self.codebuild_image = codebuild_image
            self.runtime_overrides = runtime_overrides
            self.docker_credentials_secret = dcs
            self.npm_mirror = npm_mirror
            self.npm_mirror_secret = npm_mirror_secret
            self.pypi_mirror = pypi_mirror
            self.pypi_mirror_secret = pypi_mirror_secret
            self.seedfarmer_bucket = sf_bucket

        elif self.module_manifest:
            # Placeholder for single modules
            pass



#  Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License").
#    You may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

from seedfarmer.models._base import CamelModel, ModuleRef, ValueFromRef, ValueRef
from seedfarmer.models._deploy_spec import BuildPhase, BuildPhases, BuildType, DeploySpec, ExecutionType
from seedfarmer.models._project_spec import ProjectSpec

__all__ = [
    "CamelModel",
    "ModuleRef",
    "ValueFromRef",
    "ValueRef",
    "BuildPhase",
    "BuildPhases",
    "BuildType",
    "DeploySpec",
    "ExecutionType",
    "ProjectSpec",
]



#    Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License").
#    You may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

from __future__ import annotations

from typing import Type

import seedfarmer.errors
from seedfarmer.deployment.deploy_base import DeployModule
from seedfarmer.deployment.deploy_local import DeployLocalModule
from seedfarmer.deployment.deploy_remote import DeployRemoteModule
from seedfarmer.models.transfer import ModuleDeployObject
from seedfarmer.services.session_manager import (
    ISessionManager,
    SessionManager,
    SessionManagerLocalImpl,
    SessionManagerRemoteImpl,
)


class DeployModuleFactory:
    @staticmethod
    def create(mdo: ModuleDeployObject) -> DeployModule:
        session_instance = SessionManager()
        session_type: Type[ISessionManager] = type(session_instance)
        if session_type is SessionManagerRemoteImpl:
            return DeployRemoteModule(mdo)
        elif session_type is SessionManagerLocalImpl:
            return DeployLocalModule(mdo)
        else:
            raise seedfarmer.errors.InvalidConfigurationError(f"Unsupported session type: {session_type}")

    @staticmethod
    def is_local() -> bool:
        session_instance = SessionManager()
        session_type: Type[ISessionManager] = type(session_instance)
        if session_type is SessionManagerRemoteImpl:
            return False
        else:
            return True



#  Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License").
#    You may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

import logging
import sys
from typing import Optional

import click
import yaml
from boto3 import Session

import seedfarmer.errors
import seedfarmer.mgmt.deploy_utils as du
import seedfarmer.mgmt.module_info as mi
from seedfarmer import DEBUG_LOGGING_FORMAT, config, enable_debug
from seedfarmer.output_utils import print_bolded
from seedfarmer.services.session_manager import SessionManager, bind_session_mgr

_logger: logging.Logger = logging.getLogger(__name__)


def _load_project() -> str:
    try:
        _logger.info("No --project provided, attempting load from seedfarmer.yaml")
        return config.PROJECT
    except FileNotFoundError:
        print_bolded("Unable to determine project to bootstrap, one of --project or a seedfarmer.yaml is required")
        raise click.ClickException("Failed to determine project identifier")


@click.group(name="store", help="Top Level command to support storing module information")
def store() -> None:
    "Store module information"
    pass


@store.command(
    name="deployspec",
    help="""Store/Update a deployspec of a currently deployed module.
             Use this if you cannot destroy a deployed module because of a defect
             in the the destroy portion of the deployspec.  USE WITH CAUTION as the
             existing deployspec gets overwritten and is NOT recoverable.
               """,
)
@click.option(
    "--deployment",
    "-d",
    type=str,
    help="The Deployment Name",
    required=True,
)
@click.option(
    "--group",
    "-g",
    type=str,
    help="The Group Name",
    required=True,
)
@click.option(
    "--module",
    "-m",
    type=str,
    help="The Module Name",
    required=True,
)
@click.option(
    "--path",
    type=str,
    help="""The relative module path (ex. modules/optionals/networking) -
     DO NOT PASS IN filename `deployspec.yaml`""",
    required=True,
)
@click.option(
    "--project",
    "-p",
    help="Project identifier",
    required=False,
    default=None,
)
@click.option(
    "--profile",
    default=None,
    help="The AWS profile used to create a session to assume the toolchain role",
    required=False,
)
@click.option(
    "--region",
    default=None,
    help="The AWS region used to create a session to assume the toolchain role",
    required=False,
)
@click.option(
    "--qualifier",
    default=None,
    help="""A qualifier to use with the seedfarmer roles.
     Use only if bootstrapped with this qualifier""",
    required=False,
)
@click.option(
    "--target-account-id",
    default=None,
    help="""Account Id of the target account to store deployspec, if specified --target-region is required
     You SHOULD NOT use this parameter as this command will leverage the SeedFarmer session manager!
     It is meant for development purposes.
    """,
    show_default=True,
)
@click.option(
    "--target-region",
    default=None,
    help="""Region of the target account to store deployspec, if specified --target-account-id is required
     You SHOULD NOT use this parameter as this command will leverage the SeedFarmer session manager!
     It is meant for development purposes.
    """,
    show_default=True,
)
@click.option(
    "--local/--remote",
    default=False,
    help="Indicates whether to use local session role or the SeedFarmer roles",
    show_default=True,
    type=bool,
)
@click.option(
    "--debug/--no-debug",
    default=False,
    help="Enable detailed logging.",
    show_default=True,
)
@bind_session_mgr
def store_deployspec(
    deployment: str,
    group: str,
    module: str,
    path: str,
    profile: Optional[str],
    region: Optional[str],
    qualifier: Optional[str],
    project: Optional[str],
    target_account_id: Optional[str],
    target_region: Optional[str],
    local: bool,
    debug: bool,
) -> None:
    if debug:
        enable_debug(format=DEBUG_LOGGING_FORMAT)

    _logger.debug(
        f"""Storing deployspec for module {module} of group {group}
        in deployment {deployment} located at {path}/deployspec.yaml"""
    )

    if project is None:
        project = _load_project()

    session: Session = Session(profile_name=profile, region_name=region)
    if (target_account_id is not None) != (target_region is not None):
        raise seedfarmer.errors.InvalidConfigurationError(
            "Must either specify both --target-account-id and --target-region, or neither"
        )
    elif target_account_id is not None and target_region is not None:
        session = (
            SessionManager()
            .get_or_create(project_name=project, profile=profile, region_name=region, qualifier=qualifier)
            .get_deployment_session(account_id=target_account_id, region_name=target_region)
        )

    du.update_deployspec(deployment, group, module, path, session=session)


@store.command(
    name="moduledata",
    help="""CAT or pipe in a json or yaml object.
      This command is meant to be run by seedfarmer ONLY!!!
      It is run within the context of the build job.
      Do not use this unless you are sure of the ramifications!
    """,
)
@click.option(
    "--deployment",
    "-d",
    type=str,
    help="The Deployment Name",
    required=True,
)
@click.option(
    "--group",
    "-g",
    type=str,
    help="The Group Name",
    required=True,
)
@click.option(
    "--module",
    "-m",
    type=str,
    help="The Module Name",
    required=True,
)
@click.option(
    "--project",
    "-p",
    help="Project identifier",
    required=False,
    default=None,
)
@click.option(
    "--profile",
    default=None,
    help="The AWS profile used to create a session to assume the toolchain role",
    required=False,
)
@click.option(
    "--region",
    default=None,
    help="The AWS region used to create a session to assume the toolchain role",
    required=False,
)
@click.option(
    "--qualifier",
    default=None,
    help="""A qualifier to use with the seedfarmer roles.
     Use only if bootstrapped with this qualifier""",
    required=False,
)
@click.option(
    "--target-account-id",
    default=None,
    help="Account Id of the target account to store module metadata, if specified --target-region is required",
    show_default=True,
)
@click.option(
    "--target-region",
    default=None,
    help="Region of the target account to store module metadata, if specified --target-account-id is required",
    show_default=True,
)
@click.option(
    "--local/--remote",
    default=False,
    help="Indicates whether to use local session role or the SeedFarmer roles",
    show_default=True,
    type=bool,
)
@click.option(
    "--debug/--no-debug",
    default=False,
    help="Enable detailed logging.",
    show_default=True,
)
@bind_session_mgr
def store_module_metadata(
    deployment: str,
    group: str,
    module: str,
    project: Optional[str],
    profile: Optional[str],
    region: Optional[str],
    qualifier: Optional[str],
    target_account_id: Optional[str],
    target_region: Optional[str],
    local: bool,
    debug: bool,
) -> None:
    if debug:
        enable_debug(format=DEBUG_LOGGING_FORMAT)
    _logger.debug("Writing metadata for module %s in deployment %s", module, deployment)

    if project is None:
        project = _load_project()

    session: Session = Session(profile_name=profile, region_name=region)
    if (target_account_id is not None) != (target_region is not None):
        raise seedfarmer.errors.InvalidConfigurationError(
            "Must either specify both --target-account-id and --target-region, or neither"
        )
    elif target_account_id is not None and target_region is not None:
        session = (
            SessionManager()
            .get_or_create(project_name=project, profile=profile, region_name=region, qualifier=qualifier)
            .get_deployment_session(account_id=target_account_id, region_name=target_region)
        )

    d = yaml.safe_load(sys.stdin.read())
    if d:
        mi.write_metadata(deployment=deployment, group=group, module=module, data=d, session=session)
    else:
        _logger.info("No Data avaiable...skipping")


@store.command(
    name="md5",
    help="""CAT or pipe in a string.
            This command is meant to be run by seedfarmer ONLY!!!""",
)
@click.option(
    "--deployment",
    "-d",
    type=str,
    help="The Deployment Name",
    required=True,
)
@click.option(
    "--group",
    "-g",
    type=str,
    help="The Group Name",
    required=True,
)
@click.option(
    "--module",
    "-m",
    type=str,
    help="The Module Name",
    required=True,
)
@click.option(
    "--type",
    "-t",
    type=str,
    help="The kind of MD5: bundle or spec",
    required=True,
)
@click.option(
    "--project",
    "-p",
    help="Project identifier",
    required=False,
    default=None,
)
@click.option(
    "--profile",
    default=None,
    help="The AWS profile used to create a session to assume the toolchain role",
    required=False,
)
@click.option(
    "--region",
    default=None,
    help="The AWS region used to create a session to assume the toolchain role",
    required=False,
)
@click.option(
    "--qualifier",
    default=None,
    help="""A qualifier to use with the seedfarmer roles.
     Use only if bootstrapped with this qualifier""",
    required=False,
)
@click.option(
    "--target-account-id",
    default=None,
    help="Account Id of the target account to store md5, if specified --target-region is required",
    show_default=True,
)
@click.option(
    "--target-region",
    default=None,
    help="Region of the target account to store md5, if specified --target-account-id is required",
    show_default=True,
)
@click.option(
    "--local/--remote",
    default=False,
    help="Indicates whether to use local session role or the SeedFarmer roles",
    show_default=True,
    type=bool,
)
@click.option(
    "--debug/--no-debug",
    default=False,
    help="Enable detailed logging.",
    show_default=True,
)
@bind_session_mgr
def store_module_md5(
    deployment: str,
    group: str,
    module: str,
    type: str,
    project: Optional[str],
    target_account_id: Optional[str],
    target_region: Optional[str],
    profile: Optional[str],
    region: Optional[str],
    qualifier: Optional[str],
    local: bool,
    debug: bool,
) -> None:
    if debug:
        enable_debug(format=DEBUG_LOGGING_FORMAT)
    _logger.debug("Writing md5 of  %s for module %s in deployment %s", type, module, deployment)

    if project is None:
        project = _load_project()

    session: Session = Session(profile_name=profile, region_name=region)
    if (target_account_id is not None) != (target_region is not None):
        raise seedfarmer.errors.InvalidConfigurationError(
            "Must either specify both --target-account-id and --target-region, or neither"
        )
    elif target_account_id is not None and target_region is not None:
        session = (
            SessionManager()
            .get_or_create(project_name=project, profile=profile, region_name=region, qualifier=qualifier)
            .get_deployment_session(account_id=target_account_id, region_name=target_region)
        )

    d = sys.stdin.readline().strip()
    if d:
        if type.casefold() == "bundle":
            _type = mi.ModuleConst.BUNDLE
        elif type.casefold() == "spec":
            _type = mi.ModuleConst.DEPLOYSPEC
        mi.write_module_md5(deployment=deployment, group=group, module=module, hash=d, type=_type, session=session)
    else:
        _logger.info("No Data available...skipping")



#    Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License").
#    You may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.


from __future__ import annotations

import logging
import threading
from abc import abstractmethod
from functools import update_wrapper
from threading import Thread
from time import sleep
from typing import TYPE_CHECKING, Any, Callable, Dict, List, Optional, Tuple, cast

import botocore.exceptions
from boto3 import Session
from botocore.credentials import Credentials

import seedfarmer.errors
from seedfarmer.services import boto3_client, create_new_session, create_new_session_with_creds, get_sts_identity_info
from seedfarmer.utils import get_deployment_role_arn, get_toolchain_role_arn, get_toolchain_role_name

if TYPE_CHECKING:
    from mypy_boto3_sts.type_defs import AssumeRoleResponseTypeDef

_logger: logging.Logger = logging.getLogger(__name__)


class SingletonMeta(type):
    """
    This is a thread-safe implementation of Singleton.
    """

    _instances: Dict[Any, Any] = {}

    _lock: threading.Lock = threading.Lock()

    def __call__(cls, *args: List[Any], **kwargs: Dict[Any, Any]) -> Any:
        """
        Possible changes to the value of the `__init__` argument do not affect
        the returned instance.
        """
        with cls._lock:
            if cls not in cls._instances:
                instance = super().__call__(*args, **kwargs)
                cls._instances[cls] = instance
        return cls._instances[cls]


class ISessionManager(object):
    @abstractmethod
    def get_or_create(
        self,
        *,
        project_name: Optional[str] = None,
        region_name: Optional[str] = None,
        toolchain_region: Optional[str] = None,
        qualifier: Optional[str] = None,
        role_prefix: Optional[str] = None,
        profile: Optional[str] = None,
        enable_reaper: bool = False,
        **kwargs: Optional[Any],
    ) -> Any: ...

    @property
    @abstractmethod
    def toolchain_session(self) -> Session: ...

    @abstractmethod
    def get_deployment_session(self, account_id: str, region_name: str) -> Session: ...

    @abstractmethod
    def get_toolchain_credentials(self) -> Credentials: ...


class SessionManager(ISessionManager):
    _real_instance: Optional[ISessionManager] = None

    def __new__(cls, *args: List[Any], **kwargs: Dict[str, Any]) -> SessionManager:
        if cls._real_instance is None:
            raise seedfarmer.errors.InvalidSessionError("SessionManager implementation has not been initialized")
        return cast(SessionManager, cls._real_instance)

    @classmethod
    def bind(cls, instance: ISessionManager) -> None:
        cls._real_instance = instance

    def get_or_create(self, **kwargs: Any) -> ISessionManager:
        raise NotImplementedError()

    @property
    def toolchain_session(self) -> Session:
        raise NotImplementedError()

    def get_toolchain_credentials(self) -> Credentials:
        raise NotImplementedError()

    def get_deployment_session(self, account_id: str, region_name: str) -> Session:
        raise NotImplementedError()


class SessionManagerRemoteImpl(ISessionManager, metaclass=SingletonMeta):
    TOOLCHAIN_KEY: str = "toolchain"
    SESSION: str = "session"
    ROLE: str = "role"
    sessions: Dict[str, Dict[str, Any]] = {}

    config: Dict[Any, Any] = {}
    created: bool = False
    reaper: Thread = None  # type: ignore
    reaper_interval: int = 900  # every 15 minutes

    def __init__(self) -> None:
        super().__init__()

    def get_or_create(
        self,
        *,
        project_name: Optional[str] = None,
        region_name: Optional[str] = None,
        profile: Optional[str] = None,
        toolchain_region: Optional[str] = None,
        qualifier: Optional[str] = None,
        role_prefix: Optional[str] = None,
        reaper_interval: Optional[int] = None,
        enable_reaper: bool = False,
        **kwargs: Optional[Any],
    ) -> ISessionManager:
        if not self.created:
            if project_name is None:
                raise seedfarmer.errors.InvalidConfigurationError(
                    "A 'project_name' is required for first time initialization of the SessionManager"
                )
            self.config["project_name"] = project_name
            self.config["region_name"] = region_name
            self.config["profile"] = profile
            self.config["toolchain_region"] = toolchain_region
            self.config["qualifier"] = qualifier if qualifier else None
            self.config["role_prefix"] = role_prefix if role_prefix else "/"
            self.config = {**self.config, **kwargs}
            self.toolchain_role_name = get_toolchain_role_name(project_name, cast(str, qualifier))

            if reaper_interval is not None:
                self.reaper_interval = reaper_interval

            if enable_reaper and (not self.reaper or not self.reaper.is_alive()):
                self._setup_reaper()

            self._check_for_toolchain()
            self.created = True
        return self

    @property
    def toolchain_session(self) -> Session:
        if not self.created:
            raise seedfarmer.errors.InvalidConfigurationError(
                "The SessionManager object was never properly created...)"
            )
        self._check_for_toolchain()
        return self.sessions[self.TOOLCHAIN_KEY][self.SESSION]  # type: ignore[no-any-return]

    def get_toolchain_credentials(self) -> Credentials:
        if not self.created:
            raise seedfarmer.errors.InvalidConfigurationError(
                "The SessionManager object was never properly created...)"
            )

        toolchain_role = self.sessions[self.TOOLCHAIN_KEY][self.ROLE]
        creds = Credentials(
            access_key=toolchain_role["Credentials"]["AccessKeyId"],
            secret_key=toolchain_role["Credentials"]["SecretAccessKey"],
            token=toolchain_role["Credentials"]["SessionToken"],
        )
        return creds

    def get_deployment_session(self, account_id: str, region_name: str) -> Session:
        session_key = f"{account_id}-{region_name}"
        project_name = self.config["project_name"]
        qualifier = self.config.get("qualifier") if self.config.get("qualifier") else None
        role_prefix = self.config.get("role_prefix")
        toolchain_region = self.config.get("toolchain_region")
        if not self.created:
            raise seedfarmer.errors.InvalidConfigurationError("The SessionManager object was never properly created...")
        if session_key not in self.sessions.keys():
            _logger.info(f"Creating Session for {session_key}")
            self._check_for_toolchain()
            toolchain_role = self.sessions[self.TOOLCHAIN_KEY][self.ROLE]
            # the boto sessions are not thread safe, so create a new one for the toolchain role every time to be sure
            sts_toolchain_client = boto3_client(
                service_name="sts",
                aws_access_key_id=toolchain_role["Credentials"]["AccessKeyId"],
                aws_secret_access_key=toolchain_role["Credentials"]["SecretAccessKey"],
                aws_session_token=toolchain_role["Credentials"]["SessionToken"],
                region_name=toolchain_region if toolchain_region else region_name,
            )
            partition = sts_toolchain_client.get_caller_identity()["Arn"].split(":")[1]
            deployment_role_arn = get_deployment_role_arn(
                partition=partition,
                deployment_account_id=account_id,
                project_name=project_name,
                qualifier=cast(str, qualifier),
                role_prefix=role_prefix,
            )
            _logger.debug(
                f"""The assumed toolchain role {toolchain_role["AssumedRoleUser"]["Arn"]} will
                 try and assume the deployment role: {deployment_role_arn}"""
            )
            try:
                deployment_role = sts_toolchain_client.assume_role(
                    RoleArn=deployment_role_arn,
                    RoleSessionName="deployment_role",
                )
            except botocore.exceptions.ClientError as ce:
                raise seedfarmer.errors.InvalidSessionError(
                    f"""
                {ce}
                The toolchain role cannot assume a deployment role for this account / region mapping.
                Make sure that the toolchain role is in the trust policy of the deployment role...
                   (HINT: if not, your seedfarmer bootstrap is incorrect. Use the SeedFarmer CLI to bootstrap.)
                Make sure that the account id is correct in your targetAccountMappings of the deployment manifest.
                   (HINT: look at the arn of the deployment role...the account id is REALLY important to be correct.
                   This is gotten from the deployment manifest under the targetAccountMappings section.)
                """
                )
            deployment_session = create_new_session_with_creds(
                aws_access_key_id=deployment_role["Credentials"]["AccessKeyId"],
                aws_secret_access_key=deployment_role["Credentials"]["SecretAccessKey"],
                aws_session_token=deployment_role["Credentials"]["SessionToken"],
                region_name=region_name,
            )
            self.sessions[session_key] = {self.SESSION: deployment_session, self.ROLE: deployment_role}
            return deployment_session
        else:
            return self.sessions[session_key][self.SESSION]  # type: ignore[no-any-return]

    # These methods below should not be called outside of this class

    def _check_for_toolchain(self) -> None:
        if self.TOOLCHAIN_KEY not in self.sessions.keys():
            _logger.info("Creating toolchain session")
            session, role = self._get_toolchain()
            self.sessions = {self.TOOLCHAIN_KEY: {self.SESSION: session, self.ROLE: role}}

    def _get_toolchain(self) -> Tuple[Session, "AssumeRoleResponseTypeDef"]:
        region_name = self.config.get("region_name")
        profile_name = self.config.get("profile")
        project_name = self.config.get("project_name")
        qualifier = self.config.get("qualifier") if self.config.get("qualifier") else None
        role_prefix = self.config.get("role_prefix")
        toolchain_region = self.config.get("toolchain_region")
        _logger.debug(
            f"""Creating a remote session with the following info passed in:
                      region_name = {region_name}
                      profile_name = {profile_name}
                      project_name = {project_name}
                      qualifier = {qualifier}

                      NOTE: if not set here, the active environment parameters are used to create a new AWS session
                            much like how the AWS CLI operates.
                      """
        )
        user_session = create_new_session(region_name=region_name, profile=profile_name)
        user_account_id, _, partition = get_sts_identity_info(session=user_session)
        toolchain_role_arn = get_toolchain_role_arn(
            partition=partition,
            toolchain_account_id=user_account_id,
            project_name=cast(str, project_name),
            qualifier=cast(str, qualifier),
            role_prefix=role_prefix,
        )
        _logger.debug(
            f"""The active user session will assume the toolchain role
                      arn = {toolchain_role_arn}
                      toolchain_region = {toolchain_region}
                      """
        )
        user_client = boto3_client(service_name="sts", session=user_session)
        try:
            toolchain_role = user_client.assume_role(
                RoleArn=toolchain_role_arn,
                RoleSessionName="toolchainrole",
            )
        except botocore.exceptions.ClientError as ce:
            raise seedfarmer.errors.InvalidSessionError(
                f"""
            {ce}
            The session used to call SeedFarmer is not permitted to assume the toolchain role.
            Verify the user tied to your active session is in the trust policy of the toolchain role
            or use a session that DOES have that user.
            """
            )
        toolchain_session = create_new_session_with_creds(
            aws_access_key_id=toolchain_role["Credentials"]["AccessKeyId"],
            aws_secret_access_key=toolchain_role["Credentials"]["SecretAccessKey"],
            aws_session_token=toolchain_role["Credentials"]["SessionToken"],
            region_name=toolchain_region if toolchain_region else region_name,
        )

        return toolchain_session, toolchain_role

    def _setup_reaper(self) -> None:
        _logger.info("Starting Session Reaper")
        t = Thread(target=self._reap_sessions, args=(self.reaper_interval,), daemon=True, name="SessionReaper")
        t.start()
        self.reaper = t

    def _reap_sessions(self, interval: int) -> None:
        _logger.debug("Reaper Is Set")
        while True:
            sleep(interval)
            _logger.info(f"Reaping Sessions - sleeping for {interval} seconds")
            self.sessions = {}


class SessionManagerLocalImpl(ISessionManager, metaclass=SingletonMeta):
    def __init__(self) -> None:
        self.created: bool = False
        super().__init__()

    def get_or_create(
        self,
        *,
        region_name: Optional[str] = None,
        profile: Optional[str] = None,
        **kwargs: Optional[Any],
    ) -> ISessionManager:
        if not self.created:
            # Create a session using the specified profile or default credentials
            self._session = create_new_session(region_name=region_name, profile=profile)
            self._credentials = self._session.get_credentials().get_frozen_credentials()  # type: ignore [union-attr]
            self.created = True
        return self

    @property
    def toolchain_session(self) -> Session:
        if not self.created or self._session is None:
            raise seedfarmer.errors.InvalidConfigurationError("SessionManagerLocal not properly initialized")
        return self._session

    def get_deployment_session(self, account_id: str, region_name: str) -> Session:
        # Simply return the same session regardless of account/region
        # The account_id and region are ignored...but need to meet the spec
        return self.toolchain_session

    def get_toolchain_credentials(self) -> Credentials:
        if not self.created or self._credentials is None:
            raise seedfarmer.errors.InvalidConfigurationError("SessionManagerLocal not properly initialized")
        return cast(Credentials, self._credentials)


def bind_session_mgr(f: Callable[..., Any]) -> Callable[..., Any]:
    def bind_session_mgr_inner(*args: List[Any], **kwargs: Dict[str, Any]) -> Any:
        local = kwargs.get("local", False)

        if local:
            SessionManager.bind(SessionManagerLocalImpl())
        else:
            SessionManager.bind(SessionManagerRemoteImpl())

        return f(*args, **kwargs)

    return update_wrapper(bind_session_mgr_inner, f)



#  Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License").
#    You may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

import hashlib
import logging
import os.path
import pathlib
import re
import shutil
import tarfile
from typing import Optional, Tuple
from urllib.parse import parse_qs, urlparse
from zipfile import ZipFile

import boto3
import requests
from botocore.credentials import Credentials
from requests.auth import HTTPBasicAuth
from requests.models import Response

from seedfarmer import config
from seedfarmer.errors import InvalidConfigurationError
from seedfarmer.services._secrets_manager import get_secrets_manager_value
from seedfarmer.services._service_utils import create_signed_request
from seedfarmer.services.session_manager import SessionManager

_logger: logging.Logger = logging.getLogger(__name__)


def _download_archive(archive_url: str, secret_name: Optional[str]) -> Response:
    if re.findall(r"s3\.([^\.]+\.)?amazonaws", archive_url):
        session: boto3.Session = SessionManager().get_or_create().toolchain_session
        credentials: Credentials = SessionManager().get_or_create().get_toolchain_credentials()

        signed_request = create_signed_request(
            endpoint=archive_url,
            session=session,
            credentials=credentials,
            headers={"x-amz-content-sha256": hashlib.sha256("".encode("utf-8")).hexdigest()},
        )
        return requests.get(url=signed_request.url, headers=signed_request.headers, allow_redirects=True)  # type: ignore[arg-type]

    if secret_name:
        session: boto3.Session = SessionManager().get_or_create().toolchain_session  # type: ignore[no-redef]
        secret_value = get_secrets_manager_value(secret_name, session)

        if "username" not in secret_value or "password" not in secret_value:
            raise InvalidConfigurationError(f"username and password required in secret {secret_name}")

        auth = HTTPBasicAuth(secret_value["username"], secret_value["password"])
    else:
        auth = None

    return requests.get(
        url=archive_url,
        allow_redirects=True,
        auth=auth,
    )


def _extract_archive(archive_name: str, extracted_dir_path: str) -> str:
    if archive_name.endswith(".tar.gz"):
        with tarfile.open(archive_name, "r:gz") as tar_file:
            all_members = tar_file.getmembers()
            top_level_dirs = set(member.name.split("/")[0] for member in all_members if "/" in member.name)
            if len(top_level_dirs) > 1:
                raise InvalidConfigurationError(
                    f"the archive {archive_name} can only have one directory at the root and no files"
                )
            elif len(top_level_dirs) == 1:
                embedded_dir = top_level_dirs.pop()
            else:
                embedded_dir = ""
            tar_file.extractall(extracted_dir_path, members=all_members)
    else:
        with ZipFile(archive_name, "r") as zip_file:
            all_files = zip_file.namelist()
            top_level_dirs = set(name.split("/")[0] for name in all_files if "/" in name)
            if len(top_level_dirs) > 1:
                raise InvalidConfigurationError(
                    f"the archive {archive_name} can only have one directory at the root and no files"
                )
            elif len(top_level_dirs) == 1:
                embedded_dir = top_level_dirs.pop()
            else:
                embedded_dir = ""
            for file in all_files:
                zip_file.extract(file, path=extracted_dir_path)
    return embedded_dir


def _process_archive(archive_name: str, response: Response, extracted_dir: str) -> str:
    parent_dir = os.path.join(config.OPS_ROOT, "seedfarmer.archive")
    pathlib.Path(parent_dir).mkdir(parents=True, exist_ok=True)

    with open(archive_name, "wb") as archive_file:
        archive_file.write(response.content)

    extracted_dir_path = os.path.join(parent_dir, extracted_dir)
    embedded_dir = _extract_archive(archive_name, extracted_dir_path)

    os.remove(archive_name)

    if embedded_dir:
        file_names = os.listdir(os.path.join(extracted_dir_path, embedded_dir))
        for file_name in file_names:
            shutil.move(os.path.join(extracted_dir_path, embedded_dir, file_name), extracted_dir_path)

        os.rmdir(os.path.join(extracted_dir_path, embedded_dir))

    return extracted_dir_path


def _get_release_with_link(archive_url: str, secret_name: Optional[str]) -> Tuple[str, str]:
    parent_dir = os.path.join(config.OPS_ROOT, "seedfarmer.archive")
    parsed_url = urlparse(archive_url)

    if not parsed_url.scheme == "https":
        raise InvalidConfigurationError(f"This url must be via https: {archive_url}")

    query_params = parse_qs(parsed_url.query)
    if not query_params.get("module"):
        raise InvalidConfigurationError(f"module query param required : {archive_url}")
    module = query_params["module"][0]

    archive_name = parsed_url.path.replace("/", "_")
    extracted_dir = parsed_url.path.replace(".tar.gz", "").replace(".zip", "").replace("/", "_")

    if os.path.isdir(os.path.join(parent_dir, extracted_dir)):
        return os.path.join(parent_dir, extracted_dir), module
    else:
        resp = _download_archive(
            archive_url=parsed_url._replace(fragment="", query="").geturl(),
            secret_name=secret_name,
        )

        if resp.status_code == 200:
            return _process_archive(archive_name, resp, extracted_dir), module

        else:
            _logger.error(f"Error fetching archive at {archive_url}: {resp.status_code} {resp.reason}")
            raise InvalidConfigurationError(
                f"Error fetching archive at {archive_url}: {resp.status_code} {resp.reason}"
            )


def fetch_archived_module(release_path: str, secret_name: Optional[str] = None) -> Tuple[str, str]:
    """
    Fetch an archived module from a release path. This can be a private HTTPS link.

    This function will clone the repo to the seedfarmer.archive directory
    and return the path to the cloned repo and the relative path to the module code.

    Parameters
    ----------
    release_path: str
        The path passed in to fetch. If using a ProServe provided repo, this should look like
        archive::https://github.com/awslabs/idf-modules/archive/refs/tags/v1.10.0.zip?module=modules/dummy/dummy
    secret_name: str | None
        The name of the secret to use to fetch the module repo.
        Only used if the ``release_path`` is a private HTTPS link.

    Returns
    -------
    Tuple[str,str]
        Returns a tuple that contains (in order):
            - the full path of the seedfarmer.archive where the repo was cloned to
            - the relative path to seedfarmer.archive of the module code
    """
    return _get_release_with_link(release_path.replace("archive::", ""), secret_name=secret_name)



#  Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License").
#    You may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

import datetime as dt
import logging
import os
import time
from typing import TYPE_CHECKING, Any, Callable, Dict, Optional, Tuple, Union, cast

import botocore.exceptions
from boto3 import Session

import seedfarmer.services._s3 as s3
from seedfarmer.services._service_utils import boto3_client

_logger: logging.Logger = logging.getLogger(__name__)


if TYPE_CHECKING:
    from mypy_boto3_cloudformation.waiter import StackCreateCompleteWaiter, StackUpdateCompleteWaiter

CHANGESET_PREFIX = "aws-codeseeder-"


def _wait_for_changeset(
    changeset_id: str, stack_name: str, session: Optional[Union[Callable[[], Session], Session]] = None
) -> bool:
    waiter = boto3_client("cloudformation", session=session).get_waiter("change_set_create_complete")
    try:
        waiter.wait(ChangeSetName=changeset_id, StackName=stack_name, WaiterConfig={"Delay": 1})
    except botocore.exceptions.WaiterError as ex:
        resp = ex.last_response
        status = resp["Status"]
        reason = resp["StatusReason"]
        if status == "FAILED" and (
            "The submitted information didn't contain changes." in reason or "No updates are to be performed" in reason
        ):
            _logger.debug(f"No changes for {stack_name} CloudFormation stack.")
            return False
        raise RuntimeError(f"Failed to create the changeset: {ex}. Status: {status}. Reason: {reason}")
    return True


def _create_changeset(
    stack_name: str,
    template_str: str,
    seedkit_tag: Optional[str] = None,
    template_path: str = "",
    parameters: Optional[Dict[str, str]] = None,
    session: Optional[Union[Callable[[], Session], Session]] = None,
) -> Tuple[str, str]:
    now: str = dt.datetime.now(tz=dt.timezone.utc).isoformat()
    description = f"Created by SeedFarmer at {now} UTC"
    changeset_name = CHANGESET_PREFIX + str(int(time.time()))
    stack_exist, _ = does_stack_exist(stack_name=stack_name, session=session)
    changeset_type = "UPDATE" if stack_exist else "CREATE"
    kwargs: Dict[str, Any] = {
        "ChangeSetName": changeset_name,
        "StackName": stack_name,
        "ChangeSetType": changeset_type,
        "Capabilities": ["CAPABILITY_IAM", "CAPABILITY_NAMED_IAM"],
        "Description": description,
    }
    if seedkit_tag:
        kwargs.update({"Tags": [{"Key": "codeseeder-seedkit", "Value": seedkit_tag}]})
    if template_str:
        kwargs.update({"TemplateBody": template_str})
    elif template_path:
        _logger.info(f"template_path={template_path}")
        kwargs.update({"TemplateURL": template_path})
    if parameters:
        kwargs.update({"Parameters": [{"ParameterKey": k, "ParameterValue": v} for k, v in parameters.items()]})
    resp = boto3_client("cloudformation", session=session).create_change_set(**kwargs)
    return str(resp["Id"]), changeset_type


def _execute_changeset(
    changeset_id: str, stack_name: str, session: Optional[Union[Callable[[], Session], Session]] = None
) -> None:
    boto3_client("cloudformation", session=session).execute_change_set(ChangeSetName=changeset_id, StackName=stack_name)


def _wait_for_execute(
    stack_name: str, changeset_type: str, session: Optional[Union[Callable[[], Session], Session]] = None
) -> None:
    waiter: Union["StackCreateCompleteWaiter", "StackUpdateCompleteWaiter"]
    if changeset_type == "CREATE":
        waiter = boto3_client("cloudformation", session=session).get_waiter("stack_create_complete")
    elif changeset_type == "UPDATE":
        waiter = boto3_client("cloudformation", session=session).get_waiter("stack_update_complete")
    else:
        raise RuntimeError(f"Invalid changeset type {changeset_type}")

    waiter.wait(
        StackName=stack_name,
        WaiterConfig={
            "Delay": 5,
            "MaxAttempts": 480,
        },
    )


def get_stack_name(seedkit_name: str) -> str:
    """Helper function to calculate the name of a CloudFormation Stack for a given Seedkit

    Parameters
    ----------
    seedkit_name : str
        Name of the Seedkit

    Returns
    -------
    str
        Name of the Stack Name associated with the Seedkit
    """
    return f"aws-codeseeder-{seedkit_name}"


def get_stack_status(stack_name: str, session: Optional[Union[Callable[[], Session], Session]] = None) -> str:
    """Retrieve the status of a CloudFormation Stack

    Parameters
    ----------
    stack_name : str
        Name of the CloudFormation Stack to query
    session: Optional[Union[Callable[[], Session], Session]], optional
        Optional Session or function returning a Session to use for all boto3 operations, by default None

    Returns
    -------
    str
        The status of the CloudFormation Stack, see official ``boto3`` documentation for potential status values:
        https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/cloudformation.html#CloudFormation.Client.describe_stacks

    Raises
    ------
    ValueError
        If the Stack is not found
    """
    client = boto3_client("cloudformation", session=session)
    try:
        resp = client.describe_stacks(StackName=stack_name)
        if len(resp["Stacks"]) < 1:
            raise ValueError(f"CloudFormation stack {stack_name} not found.")
    except botocore.exceptions.ClientError:
        raise
    return cast(str, resp["Stacks"][0]["StackStatus"])


def does_stack_exist(
    stack_name: str, session: Optional[Union[Callable[[], Session], Session]] = None
) -> Tuple[bool, Dict[str, str]]:
    """Checks for existence of a CloudFormation Stack while also returning Stack Outputs if it does exist

    Parameters
    ----------
    stack_name : str
        Name of the CloudFormation Stack to query
    session: Optional[Union[Callable[[], Session], Session]], optional
        Optional Session or function returning a Session to use for all boto3 operations, by default None

    Returns
    -------
    Tuple[bool, Dict[str, str]]
        Tuple2 with a boolean indicating Stack existence and a dict of any Stack Outputs
    """
    client = boto3_client("cloudformation", session=session)
    try:
        resp = client.describe_stacks(StackName=stack_name)
        if len(resp["Stacks"]) < 1:
            return (False, {})
        else:
            output = {o["OutputKey"]: o["OutputValue"] for o in resp["Stacks"][0].get("Outputs", [])}
            output["StackStatus"] = resp["Stacks"][0]["StackStatus"]
            return (True, output)
    except botocore.exceptions.ClientError as ex:
        error = ex.response["Error"]
        if error["Code"] == "ValidationError" and f"Stack with id {stack_name} does not exist" in error["Message"]:
            return (False, {})
        raise


def deploy_template(
    stack_name: str,
    filename: str,
    seedkit_tag: Optional[str] = None,
    s3_bucket: Optional[str] = None,
    parameters: Optional[Dict[str, str]] = None,
    session: Optional[Union[Callable[[], Session], Session]] = None,
) -> None:
    """Deploy a local CloudFormation Template

    The function will automatically calculate a ChangeSet if the Stack already exists and update accordingly. If the
    local template file is too large, it will be uploaded to the optional ``s3_buckeet`` and deployed from there.

    Parameters
    ----------
    stack_name : str
        Name of the CloudFormation Stack to deploy
    filename : str
        Name of the local CloudFormation template file
    seedkit_tag : Optional[str], optional
        Name of the Seedkit to Tag resources in the Stack with, by default None
    s3_bucket : Optional[str], optional
        S3 Bucket to upload the template file to if it is too large (> 51200), by default None
    parameters: Optional[Dict[str, str]], optional
        Key/Value set of Input Parameters to pass to the CloudFormation stack, by default None
    session: Optional[Union[Callable[[], Session], Session]], optional
        Optional Session or function returning a Session to use for all boto3 operations, by default None

    Raises
    ------
    FileNotFoundError
        If the local template file is not found
    ValueError
        If the S3 Bucket is not found
    """
    _logger.debug("Deploying template %s", filename)
    if not os.path.isfile(filename):
        raise FileNotFoundError(f"CloudFormation template not found at {filename}")
    template_size = os.path.getsize(filename)
    if template_size > 51_200:
        if s3_bucket is None:
            raise ValueError("s3_bucket argument is required when template size > 51 200 bytes")
        _logger.info(f"The CloudFormation template ({filename}) is too big to be deployed, using s3 bucket.")
        local_template_path = filename
        s3_file_name = filename.split("/")[-1]
        key = f"cli/remote/demo/{s3_file_name}"
        s3_template_path = f"https://s3.amazonaws.com/{s3_bucket}/{key}"
        _logger.debug("s3_template_path: %s", s3_template_path)
        s3.upload_file(src=local_template_path, bucket=s3_bucket, key=key, session=session)
        changeset_id, changeset_type = _create_changeset(
            stack_name=stack_name,
            template_str="",
            seedkit_tag=seedkit_tag,
            template_path=s3_template_path,
            parameters=parameters,
            session=session,
        )
    else:
        with open(filename, "r") as handle:
            template_str = handle.read()
        changeset_id, changeset_type = _create_changeset(
            stack_name=stack_name,
            template_str=template_str,
            seedkit_tag=seedkit_tag,
            parameters=parameters,
            session=session,
        )
    has_changes = _wait_for_changeset(changeset_id, stack_name, session=session)
    if has_changes:
        _execute_changeset(changeset_id=changeset_id, stack_name=stack_name, session=session)
        _wait_for_execute(stack_name=stack_name, changeset_type=changeset_type, session=session)


def destroy_stack(stack_name: str, session: Optional[Union[Callable[[], Session], Session]] = None) -> None:
    """Destroy the CloudFormation Stack

    Parameters
    ----------
    stack_name : str
        Name of the CloudFormation Stack to destroy
    session: Optional[Union[Callable[[], Session], Session]], optional
        Optional Session or function returning a Session to use for all boto3 operations, by default None
    """
    _logger.debug("Destroying stack %s", stack_name)
    client = boto3_client("cloudformation", session=session)
    client.delete_stack(StackName=stack_name)
    waiter = client.get_waiter("stack_delete_complete")
    waiter.wait(StackName=stack_name, WaiterConfig={"Delay": 5, "MaxAttempts": 200})



#  Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License").
#    You may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

from typing import Optional, cast

import humps
from pydantic import BaseModel, ConfigDict, model_validator

from seedfarmer.errors.seedfarmer_errors import InvalidManifestError


def to_camel(string: str) -> str:
    return cast(str, humps.camelize(string))  # type: ignore


class CamelModel(BaseModel):
    # TODO[pydantic]: The following keys were removed: `underscore_attrs_are_private`.
    # Check https://docs.pydantic.dev/dev-v2/migration/#changes-to-config for more information.
    model_config = ConfigDict(alias_generator=to_camel, populate_by_name=True, extra="allow")

    @model_validator(mode="after")
    def check_for_extra_fields(self) -> "CamelModel":
        from seedfarmer import config

        if config.MANIFEST_VALIDATION_FAIL_ON_UNKNOWN_FIELDS and self.model_extra:
            raise InvalidManifestError(f"The following keys are not allowed: {self.model_extra}")

        return self


class ModuleRef(CamelModel):
    name: str
    group: str
    key: Optional[str] = None


class ValueRef(CamelModel):
    module_metadata: Optional[ModuleRef] = None
    env_variable: Optional[str] = None
    parameter_store: Optional[str] = None
    secrets_manager: Optional[str] = None
    parameter_value: Optional[str] = None


class ValueFromRef(CamelModel):
    value_from: Optional[ValueRef] = None



#  Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License").
#    You may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

from enum import Enum
from typing import Any, List, Optional

from seedfarmer.models._base import CamelModel


class BuildType(Enum):
    BUILD_GENERAL1_SMALL = "BUILD_GENERAL1_SMALL"
    BUILD_GENERAL1_MEDIUM = "BUILD_GENERAL1_MEDIUM"
    BUILD_GENERAL1_LARGE = "BUILD_GENERAL1_LARGE"
    BUILD_GENERAL1_2XLARGE = "BUILD_GENERAL1_2XLARGE"


class BuildPhase(CamelModel):
    """
    BuildPhase
    This is a list of strings that are passed to be executed
    in their respective phases as commands
    """

    commands: List[str] = []


class BuildPhases(CamelModel):
    """
    BuildPhases
    This object has the individual commands for each of the define build phases:
    install, pre_build,  build, post_build
    """

    install: BuildPhase = BuildPhase.model_construct()
    pre_build: BuildPhase = BuildPhase.model_construct()
    build: BuildPhase = BuildPhase.model_construct()
    post_build: BuildPhase = BuildPhase.model_construct()


class ExecutionType(CamelModel):
    """
    ExecutionType
    This an object that contains the Build Phases object for the destroy or deploy
    object of the DeploySpec
    """

    phases: BuildPhases = BuildPhases.model_construct()


class DeploySpec(CamelModel):
    """
    DeploySpec
    This represents the commands passed that will be executed
    on behalf of the module to be built.
    The deploy and the destroy objects each have an ExecutionType object.
    """

    deploy: Optional[ExecutionType] = None
    destroy: Optional[ExecutionType] = None
    build_type: Optional[str] = None
    publish_generic_env_variables: Optional[bool] = True

    def __init__(self, **data: Any) -> None:
        super().__init__(**data)
        if data.get("build_type"):
            chk = str(data["build_type"]).upper()
            self.build_type = chk if chk in BuildType.__members__.keys() else BuildType.BUILD_GENERAL1_SMALL.value
        else:
            self.build_type = BuildType.BUILD_GENERAL1_SMALL.value



#  Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License").
#    You may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

import logging
import sys
from typing import Optional

import click

import seedfarmer.errors
from seedfarmer import config
from seedfarmer.mgmt import metadata_support
from seedfarmer.output_utils import print_bolded

_logger: logging.Logger = logging.getLogger(__name__)


def _load_project() -> str:
    try:
        return config.PROJECT
    except FileNotFoundError:
        print_bolded("Unable to determine project to bootstrap, seedfarmer.yaml is required")
        raise click.ClickException("Failed to determine project identifier")


@click.group(name="metadata", help="Manage the metadata in a module deployment execution")
def metadata() -> None:
    """Metadata Management group"""
    pass


@metadata.command(
    name="convert",
    help="""Convert the CDK Output of the module to SeedFarmer Metadata.
     This command is meant to be run in the deployspec only!!!
    """,
)
@click.option(
    "--jq-path",
    "-jq",
    type=str,
    help="A jq-compliant path to apply to a cdk-output (json) file",
    required=False,
)
@click.option(
    "--json-file",
    "-f",
    type=str,
    help="Relative path to a cdk-output file (defautls to cdk-exports.json)",
    required=False,
    default="cdk-exports.json",
)
def convert_cdkexports(
    json_file: str,
    jq_path: Optional[str] = None,
) -> None:
    _load_project()
    metadata_support.convert_cdkexports(jq_path=jq_path, json_file=json_file)


@metadata.command(
    name="add",
    help="""Add Output K,V to the Metadata.
     This command is meant to be run in the deployspec only!!!
    """,
)
@click.option(
    "--key",
    "-k",
    type=str,
    help="The key of a key-value pair",
    required=False,
)
@click.option(
    "--value",
    "-v",
    type=str,
    help="The value of a key-value pair",
    required=False,
)
@click.option(
    "--jsonstring",
    "-j",
    type=str,
    help="JSON-compliant string to add in a stringified format",
    required=False,
)
def add(key: str, value: str, jsonstring: str) -> None:
    _load_project()
    if jsonstring and key and value:
        raise seedfarmer.errors.InvalidConfigurationError(
            "Must either specify EITHER a json string OR a key with a value..."
        )
    if jsonstring:
        metadata_support.add_json_output(json_string=jsonstring)
    else:
        if key:
            if not value:
                raise seedfarmer.errors.InvalidConfigurationError("Must specify a key and value together")
            else:
                metadata_support.add_kv_output(key=key, value=value)
        if value and not key:
            raise seedfarmer.errors.InvalidConfigurationError("Must specify a key and value together")


@metadata.command(
    name="depmod",
    help="""Get the Full Name of the Module.
     This command is meant to be run in the deployspec only!!!
    """,
)
def depmod() -> None:
    sys.stdout.write(metadata_support.get_dep_mod_name())


@metadata.command(
    name="paramvalue",
    help="""Get the parameter value based on the suffix.
     This command is meant to be run in the deployspec only!!!
    """,
)
@click.option(
    "--suffix",
    "-s",
    type=str,
    help="A jq-compliant path to apply to a cdk-output (json) file",
    required=True,
)
def param_value(
    suffix: str,
) -> None:
    p = metadata_support.get_parameter_value(parameter_suffix=suffix)
    sys.stdout.write(p) if p else None



#  Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License").
#    You may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

import concurrent.futures
import logging
import os
from threading import Lock
from typing import Any, Dict, List, Optional, Set, Tuple, cast

import yaml
from boto3 import Session

import seedfarmer.errors
import seedfarmer.mgmt.module_info as mi
from seedfarmer.models import DeploySpec
from seedfarmer.models.manifests import DataFile, DeploymentManifest, ModuleManifest, ModulesManifest
from seedfarmer.output_utils import print_bolded
from seedfarmer.services.session_manager import SessionManager

_logger: logging.Logger = logging.getLogger(__name__)


class ModuleInfoIndex(object):
    def __init__(self) -> None:
        super().__init__()
        self._index: Dict[Tuple[str, str, str, str], Dict[str, Any]] = dict()
        self._groups: Set[str] = set()
        self._groups_idx: Dict[str, Set[Tuple[str, str, str, str]]] = dict()
        self._module_names_idx: Dict[Tuple[str, str], Tuple[str, str, str, str]] = dict()
        self._lock = Lock()

    @property
    def groups(self) -> Set[str]:
        return self._groups

    def get_module_info(
        self, *, group: str, account_id: str, region: str, module_name: str
    ) -> Optional[Dict[str, Any]]:
        return self._index.get((group, account_id, region, module_name))

    def index_module_info(
        self, *, group: str, account_id: str, region: str, module_name: str, module_info: Dict[str, Any]
    ) -> None:
        with self._lock:
            module_info_key = (group, account_id, region, module_name)
            current_module_info = self._index.get(module_info_key, {})
            self._index[module_info_key] = {**current_module_info, **module_info}
            self._groups.add(group)
            current_group_keys = self._groups_idx.get(group, set())
            current_group_keys.add(module_info_key)
            self._groups_idx[group] = current_group_keys
            self._module_names_idx[(group, module_name)] = module_info_key

    def get_keys_for_group(self, group: str) -> List[Dict[str, str]]:
        return [
            {"group": m[0], "account_id": m[1], "region": m[2], "module_name": m[3]}
            for m in self._groups_idx.get(group, [])
        ]

    def get_key_for_module_name(self, group: str, module_name: str) -> Dict[str, str]:
        m = self._module_names_idx.get((group, module_name), None)
        if m is not None:
            return {"group": m[0], "account_id": m[1], "region": m[2], "module_name": m[3]}
        else:
            return {"group": "", "account_id": "", "region": "", "module_name": ""}


def populate_module_info_index(deployment_manifest: DeploymentManifest) -> ModuleInfoIndex:
    """
    populate_module_info_index
        Fetch all info for the deployment currently stored, across all Target accounts and regions

    Parameters
    ----------
    deployment_manifest: DeploymentManifest
        The DeploymentManifest, including TargetAccount and Region mappings

    Returns
    -------
    ModuleInfoIndex
        An index of Module info for all Target accounts and regions
    """
    module_info_index = ModuleInfoIndex()

    with concurrent.futures.ThreadPoolExecutor(max_workers=len(deployment_manifest.target_accounts_regions)) as workers:

        def _get_module_info(args: Dict[str, Any]) -> None:
            role_prefix = args.pop("role_prefix", "/")
            session = (
                SessionManager()
                .get_or_create(role_prefix=role_prefix)
                .get_or_create()
                .get_deployment_session(account_id=args["account_id"], region_name=args["region"])
            )
            module_info = mi.get_parameter_data_cache(deployment=cast(str, deployment_manifest.name), session=session)
            for key, value in module_info.items():
                key_parts = key.split("/")[1:]
                if len(key_parts) < 4:
                    continue
                group = key_parts[2]
                module_name = key_parts[3]
                module_info_index.index_module_info(
                    group=group, module_name=module_name, module_info={key: value}, **args
                )

        params = [
            {
                "account_id": target_account_region["account_id"],
                "region": target_account_region["region"],
                "role_prefix": target_account_region["role_prefix"],
            }
            for target_account_region in deployment_manifest.target_accounts_regions
        ]
        _ = list(workers.map(_get_module_info, params))

    return module_info_index


def validate_data_files(data_files: Optional[List[DataFile]]) -> List[str]:
    """
    validate_data_files
        This will determine if all data files requested are available, and return the files that are not available.

    Parameters
    ----------
    data_files : Optional[List[DataFile]]
        The list of DataFile objects to evaluate

    Returns
    -------
    List[str]
        The list of data file paths not found
    """
    missing_files = []
    if data_files is not None:
        missing_files = [
            data_file.file_path for data_file in data_files if not os.path.isfile(str(data_file.get_local_file_path()))
        ]
    return missing_files


def validate_group_parameters(group: ModulesManifest) -> None:
    """
    validate_group_parameters
        This will verify that there are no intra-group dependencies in the parameter references

    Parameters
    ----------
    group: ModulesManifest
        The ModulesManifest representing a group

    Returns
    -------
    None
    """
    _logger.debug(f"Inspecting group {group.name} for intra-dependencies")
    group_wip = set({})
    for module in group.modules:
        if module.parameters:
            for parameter in module.parameters:
                (
                    group_wip.add(parameter.value_from.module_metadata.group)
                    if parameter.value_from and parameter.value_from.module_metadata
                    else None
                )
    if group.name in group_wip:
        message = f"""
        ERROR!!!  An intra-group dependency for was found in a module reference for group {group.name}
          No module can refer to its own group for parameter lookups!!
        """
        print_bolded(message=message, color="red")
        raise seedfarmer.errors.InvalidConfigurationError(message)


def validate_module_dependencies(
    module_dependencies: Dict[str, List[str]], destroy_manifest: DeploymentManifest
) -> List[Dict[str, List[str]]]:
    """
    This will compare a dictionary of the module dependencies and a destroy manifest object to make sure there
    are no modules scheduled to be destroyed that are referenced by modules that will be / are deployed

    Parameters
    ----------
    module_dependencies : Dict[str,List]
        A dict that is that has the module as the key (in form of `<group>-<module_name>`) and the value is a list
        of modules (in form of `[<group>-<module_name>]`) that are dependent on that module
    destroy_manifest : DeploymentManifest
        The manifest object representing all modules that should be destroyed

    Returns
    -------
    List[Dict[str,List[str]]]
        A list of dictionaries of the modules that have dependencies that are blocking deletion:
        ```
        [
            {<group>-<module_name>:[<group>-<module_name>,<group>-<module_name>]}
        ]
        ```

    """

    def _get_module_list(manifest: DeploymentManifest) -> List[str]:
        module_list = []
        for group in manifest.groups:
            for module in group.modules:
                module_list.append(f"{group.name}-{module.name}")
        return module_list

    volations = []
    module_destroy_list = _get_module_list(destroy_manifest)
    for destroy_mod_candidate in module_destroy_list:
        mod_dep = (
            module_dependencies.get(destroy_mod_candidate) if module_dependencies.get(destroy_mod_candidate) else None
        )
        if mod_dep:
            v = [module for module in mod_dep if module not in module_destroy_list]
            violation = {destroy_mod_candidate: v}
            volations.append(violation)

    return volations


def generate_dependency_maps(manifest: DeploymentManifest) -> Tuple[Dict[str, List[str]], Dict[str, List[str]]]:
    """
    Takes a deployment manifest object and returns two (2) dictionaries that contain:
        1. all the other modules that a given module depends on (module_depends_on)
        2. all the modules that are dependent on a given module (module_dependencies)

    Parameters
    ----------
    manifest : DeploymentManifest
        A fully populated deployment manifest object

    Returns
    -------
    Tuple[Dict[str, List], Dict[str, List]]
        2 Dictionaries:
            1. `module_depends_on` - all the other modules that a given module depends on
            2. `module_dependencies` - all the modules that are dependent on the given module
    """

    def add_to_list(target_dict: Dict[str, Any], key: str, val: str) -> None:
        active_list = target_dict.get(key) if target_dict.get(key) else []
        active_list.append(val) if val not in active_list else None  # type: ignore
        target_dict[key] = active_list

    module_depends_on: Dict[str, Any] = {}
    module_dependencies: Dict[str, Any] = {}
    for group in manifest.groups:
        for module in group.modules:
            group_module_name = f"{group.name}-{module.name}"
            for parameter in module.parameters:
                if parameter.value_from and parameter.value_from.module_metadata:
                    parameter_module_reference = (
                        f"{parameter.value_from.module_metadata.group}-{parameter.value_from.module_metadata.name}"
                    )
                    add_to_list(module_depends_on, group_module_name, parameter_module_reference)
                    add_to_list(module_dependencies, parameter_module_reference, group_module_name)
    return module_depends_on, module_dependencies


def prepare_ssm_for_deploy(
    deployment_name: str, group_name: str, module_manifest: ModuleManifest, account_id: str, region: str
) -> None:
    """
    prepare_ssm_for_deploy
        This method takes the populated ModuleManifest and updates SSM to prepare for deployment

    Parameters
    ----------
    deployment_name : str
        The deployment name
    group_name : str
        The group name
    module_manifest : ModuleManifest
        The Module Manifect opject
    account_id : str
        The Account Id of where this module is to be deployed
    region : str
        The Region of where this module is to be deployed
    """

    # Remove the deployspec before writing...remove bloat as we write deployspec separately
    session = SessionManager().get_or_create().get_deployment_session(account_id=account_id, region_name=region)
    module_manifest_wip = module_manifest.model_copy()
    module_manifest_wip.deploy_spec = None
    mi.write_module_manifest(
        deployment=deployment_name,
        group=group_name,
        module=module_manifest.name,
        data=module_manifest_wip.model_dump(),
        session=session,
    )
    (
        mi.write_deployspec(
            deployment=deployment_name,
            group=group_name,
            module=module_manifest.name,
            data=module_manifest.deploy_spec.model_dump(),
            session=session,
        )
        if module_manifest.deploy_spec
        else None
    )
    (
        mi.write_module_md5(
            deployment=deployment_name,
            group=group_name,
            module=module_manifest.name,
            hash=module_manifest.deployspec_md5,
            type=mi.ModuleConst.DEPLOYSPEC,
            session=session,
        )
        if module_manifest.deployspec_md5
        else None
    )
    (
        mi.write_module_md5(
            deployment=deployment_name,
            group=group_name,
            module=module_manifest.name,
            hash=module_manifest.manifest_md5,
            type=mi.ModuleConst.MANIFEST,
            session=session,
        )
        if module_manifest.manifest_md5
        else None
    )
    mi.remove_module_md5(
        deployment=deployment_name,
        group=group_name,
        module=module_manifest.name,
        type=mi.ModuleConst.BUNDLE,
        session=session,
    )


def write_deployed_deployment_manifest(deployment_manifest: DeploymentManifest) -> None:
    """
    write_deployed_deployment_manifest
        Write the deployment manifest to the store

    Parameters
    ----------
    deployment_manifest : DeploymentManifest
        The deployment manifest ojject to store
    """
    deployment_name = cast(str, deployment_manifest.name)
    for group in deployment_manifest.groups:
        delattr(group, "modules")
    session = SessionManager().get_or_create().toolchain_session
    mi.write_deployed_deployment_manifest(
        deployment=deployment_name, data=deployment_manifest.model_dump(), session=session
    )


def generate_deployed_manifest(
    deployment_name: str,
    skip_deploy_spec: bool = False,
    ignore_deployed: Optional[bool] = False,
) -> Optional[DeploymentManifest]:
    """
    Generate a DeploymentManifest object from based off deployed modules in a deployment

    Parameters
    ----------
    deployment_name : str
        The name of the deployment the manifest is generated for
    skip_deploy_spec : bool, optional
        Skip populating each module deployspec - handy for getting the list of deployed modules quickly
    ignore_deployed : Optional[bool]
        When fetching the deployment manifest stored, ignore the successfully deployed modules,
        forcing a fetch of the last requested deployment (to include modules that failed to deploy)

    Returns
    -------
    Optional[DeploymentManifest]
        The hydrated DeploymentManifest object of deployed modules
    """
    session_manager = SessionManager().get_or_create()
    dep_manifest_dict = mi.get_deployed_deployment_manifest(deployment_name, session=session_manager.toolchain_session)
    if dep_manifest_dict is None or ignore_deployed:
        # No successful deployments, just use what was last requested
        dep_manifest_dict = mi.get_deployment_manifest(deployment_name, session=session_manager.toolchain_session)
    deployed_manifest = None
    if dep_manifest_dict:
        deployed_manifest = DeploymentManifest(**dep_manifest_dict)
        module_info_index = populate_module_info_index(deployment_manifest=deployed_manifest)
        for module_group in dep_manifest_dict["groups"] if dep_manifest_dict["groups"] else []:
            group_name = module_group["name"]
            module_group["modules"] = _populate_group_modules_from_index(
                deployment_name=deployment_name,
                group_name=group_name,
                module_info_index=module_info_index,
                skip_deploy_spec=skip_deploy_spec,
            )
        deployed_manifest.groups = [ModulesManifest(**group) for group in dep_manifest_dict["groups"]]
    return deployed_manifest


def get_deployed_group_ordering(deployment_name: str) -> Dict[str, int]:
    """
    This generates a dict of the groups deployed and the index representing the proper deployment ordering

    Parameters
    ----------
    deployment_name : str
        The name of the deployment

    Returns
    -------
    Dict[str, int]
        A dict with the name of the group as the key and an int as the value of the index
    """
    session_manager = SessionManager().get_or_create()
    dep_manifest_dict = mi.get_deployed_deployment_manifest(deployment_name, session=session_manager.toolchain_session)
    if dep_manifest_dict is None:
        # No successful deployments, just use what was last requested
        dep_manifest_dict = mi.get_deployment_manifest(deployment_name, session=session_manager.toolchain_session)
    ordering = {}
    if dep_manifest_dict and dep_manifest_dict.get("groups"):
        for idx, val in enumerate(dep_manifest_dict["groups"]):
            ordering[val["name"]] = idx
    return ordering


def force_redeploy(
    group_name: str, module_name: str, active_modules: List[str], module_upstream_dep: Dict[str, List[str]]
) -> bool:
    """
    force_redeploy
    This indicates whether a module needs to redeployed based on a list of other modules that have been scheduled
    to be redeployed.

    Parameters
    ----------
    group_name : str
        The name of the group the module belongs to
    module_name : str
        The name of the modules
    active_modules : List[str]
        A list of modules already scheduled to be deployed.  The have the naming format of <group_name>-<module_name>
    module_upstream_dep : Dict[str, List[str]]
        A dict containing all the upstream dependencies of a module.  Each key in the dict is a module name
        with the format <group_name>-<module_name> and the value is a list of modules, each with the format
        of <group_name>-<module_name>
    Returns
    -------
    bool
        Indicator whether or not the module needs to be force-redeployed.
    """
    if module_upstream_dep.get(f"{group_name}-{module_name}"):
        return (
            True
            if (len((set(active_modules)).intersection(set(module_upstream_dep[f"{group_name}-{module_name}"]))) > 0)
            else False
        )
    else:
        return False


def need_to_build(
    deployment_name: str,
    group_name: str,
    module_manifest: ModuleManifest,
    active_modules: List[str] = [],
    module_upstream_dep: Dict[str, List[str]] = {},
    force_redeploy_flag: bool = False,
    deployment_params_cache: Optional[Dict[str, Any]] = None,
) -> bool:
    """
    This will indicated whether a module needs to be rebuilt based on
    the manifests, deployspec, and code of the groups and modules.  It will also
    record this info if the module does need to be rebuilt.

    Parameters
    ----------
    deployment_name : str
        The name of the deployment
    group_name : str
        A name of the corresponding group
    module_manifest : ModuleManifest
        The populated ModuleManifest Object
    active_modules : List[str]
        A list of modules already scheduled to be deployed.  The have the naming format of <group_name>-<module_name>
    module_upstream_dep : Dict[str, List[str]]
        A dict containing all the upstream dependencies of a module.  Each key in the dict is a module name
        with the format <group_name>-<module_name> and the value is a list of modules, each with the format
        of <group_name>-<module_name>
    force_redeploy_flag: bool
        A bool indicating whether the deployment is set to force-redeploy
        Defaults to False
    deployment_params_cache: Dict[str, Any]
        A cache of deployment commands

    Returns
    -------
    bool
        Indicator whether the module needs to be build or not.
        True - yes, built it
        False - no, do not build it
    """
    if force_redeploy_flag and (
        force_redeploy(
            group_name=group_name,
            module_name=module_manifest.name,
            active_modules=active_modules,
            module_upstream_dep=module_upstream_dep,
        )
    ):
        return True

    session = (
        SessionManager()
        .get_or_create()
        .get_deployment_session(
            account_id=cast(str, module_manifest.get_target_account_id()),
            region_name=cast(str, module_manifest.target_region),
        )
    )
    if (
        mi.does_md5_match(
            deployment_name,
            group_name,
            module_manifest.name,
            module_manifest.bundle_md5,  # type: ignore
            mi.ModuleConst.BUNDLE,
            deployment_params_cache,
            session=session,
        )
        and mi.does_md5_match(
            deployment_name,
            group_name,
            module_manifest.name,
            module_manifest.deployspec_md5,  # type: ignore
            mi.ModuleConst.DEPLOYSPEC,
            deployment_params_cache,
            session=session,
        )
        and mi.does_md5_match(
            deployment_name,
            group_name,
            module_manifest.name,
            module_manifest.manifest_md5,  # type: ignore
            mi.ModuleConst.MANIFEST,
            deployment_params_cache,
            session=session,
        )
    ):
        return False
    else:
        return True


def write_group_manifest(deployment_name: str, group_manifest: ModulesManifest) -> None:
    """
    write_group_manifest
        Persist the group from a MOdulesManifest object

    Parameters
    ----------
    deployment_name : str
        The name of the deployment
    group_manifest : ModulesManifest
        The ModulesManifest object of the groups to persist
    """
    g = group_manifest.name if group_manifest.name else ""
    mi.write_group_manifest(deployment=deployment_name, group=g, data=group_manifest.model_dump())


def filter_deploy_destroy(apply_manifest: DeploymentManifest, module_info_index: ModuleInfoIndex) -> DeploymentManifest:
    """
    This method takes a populated DeploymentManifest object based off the requested deployment.
    It compares this requested deployment with what is currently deployed.  If there are groups or
    modules that are not in the request, a DeploymentManifest object is returned populated with all
    the information of modules to be destroyed.

    Parameters
    ----------
    apply_manifest : DeploymentManifest
        The DeploymentManifest object based off a deployment manifest of a requested deployment.
    module_info_index: ModuleInfoIndex
        The index of existing module info for all Target accounts and regions

    Returns
    -------
    DeploymentManifest
        A populated DeploymentManifest object with the modules needed to be destroyed
    """
    deployment_name = cast(str, apply_manifest.name)

    destroy_manifest = apply_manifest.model_copy()
    delattr(destroy_manifest, "groups")
    destroy_group_list = _populate_groups_to_remove(deployment_name, apply_manifest.groups, module_info_index)
    destroy_manifest.groups = destroy_group_list
    for group in apply_manifest.groups:
        destroy_module_list = _populate_modules_to_remove(deployment_name, group.name, group.modules, module_info_index)
        if destroy_module_list:
            to_destroy = ModulesManifest(name=group.name, path=group.path, modules=destroy_module_list)
            destroy_manifest.groups.append(to_destroy)

    # Make sure the groups are sorted in proper order of the deployment
    try:
        ordering = get_deployed_group_ordering(deployment_name)

        def groupOrderingFilter(module: ModulesManifest) -> int:
            return ordering.get(module.name, 99)

        destroy_manifest.groups.sort(key=groupOrderingFilter)
    except Exception as e:
        _logger.info(f"Threw and error trying to sort the groups for destroy, ignoring and moving on {e}")

    destroy_manifest.validate_and_set_module_defaults()
    return destroy_manifest


def _populate_group_modules_from_index(
    deployment_name: str, group_name: str, module_info_index: ModuleInfoIndex, skip_deploy_spec: bool
) -> List[Dict[str, Any]]:
    modules = []
    for group_key in module_info_index.get_keys_for_group(group_name):
        deployment_params_cache = module_info_index.get_module_info(**group_key)
        module_manifest = mi.get_module_manifest(
            deployment_name, group_name, group_key["module_name"], deployment_params_cache
        )
        if module_manifest is not None:
            if not skip_deploy_spec:
                module_manifest["deploy_spec"] = mi.get_deployspec(
                    deployment_name, group_name, group_key["module_name"], deployment_params_cache
                )
            modules.append(module_manifest)
    return modules


def _populate_groups_to_remove(
    deployment_name: str, apply_groups: List[ModulesManifest], module_info_index: ModuleInfoIndex
) -> List[ModulesManifest]:
    set_requested_groups = set([g.name for g in apply_groups])
    set_deployed_groups = module_info_index.groups
    destroy_groups = list(sorted(set_deployed_groups - set_requested_groups))
    _logger.debug("Groups already deployed that will be DESTROYED : %s", destroy_groups)
    destroy_group_list = []
    for destroy_group in destroy_groups:
        group_manifest = {
            "name": destroy_group,
            "modules": _populate_group_modules_from_index(
                deployment_name=deployment_name,
                group_name=destroy_group,
                module_info_index=module_info_index,
                skip_deploy_spec=False,
            ),
        }
        destroy_group_list.append(ModulesManifest(**group_manifest))  # type: ignore[arg-type]
    return destroy_group_list


def _populate_modules_to_remove(
    deployment_name: str, group: str, apply_modules: List[ModuleManifest], module_info_index: ModuleInfoIndex
) -> List[ModuleManifest]:
    set_requests_modules = set([m.name for m in apply_modules])
    set_deployed_modules = set([m["module_name"] for m in module_info_index.get_keys_for_group(group)])
    destroy_modules = list(sorted(set_deployed_modules - set_requests_modules))
    _logger.debug("Modules of Group %s already deployed that will be DESTROYED: %s", group, destroy_modules)

    destroy_module_list = []
    for destroy_module in destroy_modules:
        deployment_params_cache = module_info_index.get_module_info(
            **module_info_index.get_key_for_module_name(group=group, module_name=destroy_module)
        )
        module_manifest = mi.get_module_manifest(deployment_name, group, destroy_module, deployment_params_cache)
        if module_manifest is not None:
            module_manifest["deploy_spec"] = mi.get_deployspec(
                deployment_name, group, destroy_module, deployment_params_cache
            )
            destroy_module_list.append(ModuleManifest(**module_manifest))
    return destroy_module_list


def update_deployspec(
    deployment: str, group: str, module: str, module_path: str, session: Optional[Session] = None
) -> None:
    d_path = mi.get_deployspec_path(module_path=module_path)
    with open(d_path) as deploymentspec:
        new_spec = DeploySpec(**yaml.safe_load(deploymentspec))
    mi.write_deployspec(deployment, group, module, new_spec.model_dump(), session=session)



#  Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License").
#    You may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

import logging
import os
from pathlib import Path
from typing import Optional

import yaml
from cookiecutter.main import cookiecutter

from seedfarmer import config

_logger: logging.Logger = logging.getLogger(__name__)


def remove_leading_path_sep(path: str) -> str:
    if path.startswith(os.sep):
        return str(path[len(os.sep) :])
    # On Windows, also check for altsep ('/')
    if os.altsep and path.startswith(os.altsep):
        return str(path[len(os.altsep) :])
    return path


def add_module_manifest(module_name: str, module_path: str) -> None:
    """Add the module manifest to the project.

    Creates the module manifest and adds the path to the groups in deployment.yaml

    Parameters
    ----------
    module_name : str
        Name of the module
    module_path: str
        The absolute path of the module location...assumed that seedfarmer.yaml
        has determined this pat
    """

    module_path = remove_leading_path_sep(module_path.replace(config.OPS_ROOT, ""))
    _logger.debug("module_path=%s", module_path)
    _logger.debug("module_name=%s", module_name)

    module_manifest = {
        "name": module_name,
        "path": module_path,
    }
    manifest_name = f"{module_name}-group.yml"
    manifest_path = os.path.join(config.OPS_ROOT, manifest_name)
    with open(manifest_path, "w") as f:
        yaml.dump(module_manifest, f)

    group_definition = {"name": f"{module_name}-group", "path": manifest_name}
    deployment_yaml = os.path.join(config.OPS_ROOT, "deployment.yaml")
    # Check if the file exists
    if os.path.exists(deployment_yaml):
        with open(deployment_yaml, "r") as file:
            data = yaml.safe_load(file) or {}
        if "groups" in data:
            if isinstance(data["groups"], list):
                data["groups"].append(group_definition)
            else:
                data["groups"] = [data["groups"], group_definition]
        else:
            data["groups"] = [group_definition]

        with open(deployment_yaml, "w") as file:
            yaml.safe_dump(data, file)
    else:
        _logger.info(f"YAML file does not exist at path: {deployment_yaml}, skipping append")


def create_module_dir(
    module_name: str,
    module_type: Optional[str],
    template_url: Optional[str],
    group_name: Optional[str] = None,
    template_branch: Optional[str] = "init-module",
) -> None:
    """Initializes a directory for a new module.

    Creates a new directory that contains files that will aid in setting up a development environment

    Parameters
    ----------
    group_name : Optional[str],
        Name of the group where the module will reside. If group is a nested dir, use `/` as a delimiter
    module_name : str
        Name of the module. The initialization will include project files pulled from the template_url
    module_type: Optional[str]
        They type of code the module deploys with, adding more boilerplate code
        -- only cdkv2 is supported here
    template_url : Optional[List[str]]
        A URL, for example a Github repo, that is or contains the template for the for the initialization
    template_branch : Optional[str]
        The Branch on the template repository. If not specified, the default template branch is `main`
    """
    module_name = module_name.replace("_", "-")
    module_root = os.path.join(config.OPS_ROOT, "modules")
    module_path = os.path.join(module_root, module_name)
    output_dir = module_root
    checkout_branch = template_branch

    module_path = (
        os.path.join(module_root, group_name, module_name) if group_name else os.path.join(module_root, module_name)
    )
    output_dir = os.path.join(module_root, group_name) if group_name else os.path.join(module_root)

    if not os.path.exists(output_dir):
        _logger.info("Creating group dir: %s", output_dir)
        os.makedirs(output_dir)

    if not os.path.exists(module_path):
        if template_url == "https://github.com/awslabs/seed-farmer.git":
            checkout_branch = "init-module-cdkv2" if module_type == "cdkv2" else "init-module"
        if template_branch is not None:
            checkout_branch = template_branch

        _logger.info("New module will be created in the following dir: %s", output_dir)
        cookiecutter(
            template=template_url,
            checkout=checkout_branch,
            no_input=True,
            extra_context={"project_name": module_name, "module_name": module_name},
            output_dir=output_dir,
        )

        add_module_manifest(module_name, module_path)


def create_project(
    template_url: str = "https://github.com/awslabs/seed-farmer.git",
    template_branch: Optional[str] = "init-project",
    project_name: Optional[str] = None,
    project_dir: Optional[str] = None,
) -> None:
    """Initializes a new project directory.

    Creates a new directory that contains files that will aid in setting up a development environment

    Parameters
    ----------
    template_url : str
        A URL, for example a Github repo, that is or contains the template for the initialization
    template_branch : Optional[str]
        The Branch on the template repository. If not specified, the default template branch is `main`
    project_name : Optional[str]
        Name of the project. The initialization will include project files pulled from the template_url
        using the template_branch as reference
    project_root_name: Optional[str]
        If the name of the directory is not to be ths same name of the project, override it here
    """
    # Create minimal seedfarmer.yaml first if project_name is provided
    if project_name is not None:
        with open(os.path.join(os.getcwd(), "seedfarmer.yaml"), "w") as f:
            yaml.dump({"project": project_name}, f, default_flow_style=False)

    project_name = project_name if project_name else config.PROJECT
    project_dir = project_dir if project_dir else project_name

    proposed_project_path = Path(project_dir)

    if proposed_project_path.is_dir():
        print("The targeted project / capability directory already exists. Exiting.")
        exit(0)

    cookiecutter(
        template=template_url,
        checkout=template_branch,
        no_input=True,
        extra_context={"project_name": project_name, "project_slug": project_dir},
        output_dir=config.OPS_ROOT,
    )

    if project_name is not None:
        # Move the seedfarmer.yaml to the project directory
        os.replace(
            os.path.join(os.getcwd(), "seedfarmer.yaml"), os.path.join(config.OPS_ROOT, project_dir, "seedfarmer.yaml")
        )
    else:
        os.replace(
            os.path.join(config.OPS_ROOT, config.CONFIG_FILE),
            os.path.join(config.OPS_ROOT, config.PROJECT, config.CONFIG_FILE),
        )



#  Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License").
#    You may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

from seedfarmer.services._service_utils import (
    boto3_client,
    boto3_resource,
    create_new_session,
    create_new_session_with_creds,
    get_botocore_config,
    get_region,
    get_sts_identity_info,
)

__all__ = [
    "get_botocore_config",
    "get_account_id",
    "get_region",
    "boto3_client",
    "boto3_resource",
    "create_new_session",
    "create_new_session_with_creds",
    "get_sts_identity_info",
]



#  Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License").
#    You may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

import json
import logging
import os
import sys
from enum import Enum
from typing import Any, Dict, List, Optional, Tuple

from boto3 import Session

import seedfarmer.errors
from seedfarmer import config
from seedfarmer.services import _secrets_manager as secrets
from seedfarmer.services import _ssm as ssm
from seedfarmer.utils import generate_hash, generate_session_hash, remove_nulls

_logger: logging.Logger = logging.getLogger(__name__)


class ModuleConst(Enum):
    DEPLOYSPEC = "deployspec"
    BUNDLE = "bundle"
    METADATA = "metadata"
    MD5 = "md5"
    MANIFEST = "manifest"
    DEPLOYED = "deployed"


def get_parameter_data_cache(deployment: str, session: Session) -> Dict[str, Any]:
    """
    get_parameter_data_cache
        Fetch the deployment parameters stored

    Parameters
    ----------
    deployment : str
        Name of the deployment
    session: Session, optional
        The boto3.Session to use to for SSM Parameter queries, default None

    Returns
    -------
    Dict[str,Any]
        A dictionary representation of what is in the store (SSM for DDB) of the modules deployed
    """
    return ssm.get_all_parameter_data_by_path(prefix=_deployment_key(deployment), session=session)


def get_all_deployments(session: Optional[Session] = None) -> List[str]:
    """
    get_all_deployments
        Get all names of curently deployments
    session: Session, optional
        The boto3.Session to use to for SSM Parameter queries, default None
    session: Session, optional
        The boto3.Session to use to for SSM Parameter queries, default None

    Returns
    -------
    List[str]
        A list of the deployments in the account
    """
    prefix = f"/{config.PROJECT}"
    _filter = f"{ModuleConst.MANIFEST.value}"
    ret = set()
    params = ssm.list_parameters_with_filter(prefix=prefix, contains_string=_filter, session=session)
    for param in params:
        _logger.debug(param)
        p = param.split("/")[3]
        if ModuleConst.MANIFEST.value in p:
            ret.add(param.split("/")[2])
    return list(ret)


def get_all_groups(
    deployment: str, params_cache: Optional[Dict[str, Any]] = None, session: Optional[Session] = None
) -> List[str]:
    """
    get_all_groups
        Get all groups in a deployment

    Parameters
    ----------
    deployment : str
        The name of the deployment
    params_cache : Dict[str,Any], optional
        A populated dict with the key  of the parameter stored and its value
    session: Session, optional
        The boto3.Session to use to for SSM Parameter queries, default None

    Returns
    -------
    List[str]
        A list of the group names
    """
    prefix = _deployment_key(deployment)
    _filter = f"{ModuleConst.MANIFEST.value}"
    ret = set()
    params = (
        params_cache.keys()
        if params_cache
        else ssm.list_parameters_with_filter(prefix=prefix, contains_string=_filter, session=session)
    )
    for param in params:
        p = param.split("/")[3]
        if ModuleConst.MANIFEST.value not in p:
            ret.add(param.split("/")[3])
    return list(ret)


def get_deployed_modules(
    deployment: str, group: str, params_cache: Optional[Dict[str, Any]] = None, session: Optional[Session] = None
) -> List[str]:
    """
    get_deployed_modules
        Get all modules deployed in a group of a deployment

    Parameters
    ----------
    deployment : str
        The name of the deployment
    group : str
        The name of the group
    params_cache : Dict[str,Any], optional
        A populated dict with the key  of the parameter stored and its value
    session: Session, optional
        The boto3.Session to use to for SSM Parameter queries, default None

    Returns
    -------
    List[str]
        A list of the names of the modules in the group
    """
    prefix = f"/{config.PROJECT}/{deployment}/{group}"
    _filter = f"{ModuleConst.MD5.value}/{ModuleConst.BUNDLE.value}"
    params = params_cache.keys() if params_cache else ssm.list_parameters_with_filter(prefix, _filter, session=session)
    ret: List[str] = []
    for param in params:
        ret.append(param.split("/")[4]) if _filter in param else None
    return ret


def get_module_md5(
    deployment: str, group: str, module: str, type: ModuleConst, session: Optional[Session] = None
) -> Optional[str]:
    """
    get_module_md5
        Get the md5 of a currently deployed module.
        There are three (3) relative md5 hashes of the module:
            - bundle / zip of currently deployed module
            - deployspec of currently deployed module
            - manifest of currently deployed module

    Parameters
    ----------
    deployment : str
        The name of the deployment
    group : str
        The name of the group
    module : str
        The name of the module
    type : ModuleConst
        An emumeration value of the md5 of the module you want
    session: Session, optional
        The boto3.Session to use to for SSM Parameter queries, default None

    Returns
    -------
    Optional[str]
        The md5 hash as a string
    """
    name = _md5_module_key(deployment, group, module, type)
    p = ssm.get_parameter_if_exists(name=name, session=session)
    return p["hash"] if p else None


def get_module_metadata(
    deployment: str,
    group: str,
    module: str,
    params_cache: Optional[Dict[str, Any]] = None,
    session: Optional[Session] = None,
) -> Optional[Dict[str, Any]]:
    """
    get_module_metadata
        Get the metadata stored for a deployed module

    Parameters
    ----------
    deployment : str
        The name of the deployment
    group : str
        The name of the group
    module : str
        The name of the module
    params_cache : Dict[str,Any], optional
        A populated dict with the key  of the parameter stored and its value
    session: Session, optional
        The boto3.Session to use to for SSM Parameter queries, default None

    Returns
    -------
    Optional[Dict[str, Any]]
        A dict containg the metadata of the module requested
    """
    return _fetch_helper(_metadata_key(deployment, group, module), params_cache, session=session)


def get_module_manifest(
    deployment: str,
    group: str,
    module: str,
    params_cache: Optional[Dict[str, Any]] = None,
    session: Optional[Session] = None,
) -> Optional[Dict[str, Any]]:
    """
    get_module_manifest
        Get the manifest stored for a deployed module

    Parameters
    ----------
    deployment : str
        The name of the deployment
    group : str
        The name of the group
    module : str
        The name of the module
    params_cache : Dict[str,Any], optional
        A populated dict with the key  of the parameter stored and its value
    session: Session, optional
        The boto3.Session to use to for SSM Parameter queries, default None

    Returns
    -------
    Optional[Dict[str, Any]]
        A dict containg the manifest of the module requested
    """
    return _fetch_helper(_manifest_key(deployment, group, module), params_cache, session=session)


def get_deployspec(
    deployment: str,
    group: str,
    module: str,
    params_cache: Optional[Dict[str, Any]] = None,
    session: Optional[Session] = None,
) -> Optional[Dict[str, Any]]:
    """
    get_deployspec
        Get the deployspec stored for a deployed module

    Parameters
    ----------
    deployment : str
        The name of the deployment
    group : str
        The name of the group
    module : str
        The name of the module
    params_cache : Dict[str,Any], optional
        A populated dict with the key  of the parameter stored and its value
    session: Session, optional
        The boto3.Session to use to for SSM Parameter queries, default None

    Returns
    -------
    Optional[Dict[str, Any]]
        A dict containg the deployspec of the module requested
    """
    return _fetch_helper(_deployspec_key(deployment, group, module), params_cache, session=session)


def get_deployment_manifest(
    deployment: str, params_cache: Optional[Dict[str, Any]] = None, session: Optional[Session] = None
) -> Optional[Dict[str, Any]]:
    """
    get_deployment_manifest
        Get the deployment manifest stored for a deployment

    Parameters
    ----------
    deployment : str
        The name of the deployment
    params_cache : Dict[str,Any], optional
        A populated dict with the key  of the parameter stored and its value
    session: Session, optional
        The boto3.Session to use to for SSM Parameter queries, default None

    Returns
    -------
    Optional[Dict[str, Any]]
        A dict of the deployment
    """
    return _fetch_helper(_deployment_manifest_key(deployment), params_cache, session=session)


def get_deployed_deployment_manifest(
    deployment: str, params_cache: Optional[Dict[str, Any]] = None, session: Optional[Session] = None
) -> Optional[Dict[str, Any]]:
    """
    get_deployed_deployment_manifest
        Get the deployment manifest stored for a completed successful deployment deployment

    Parameters
    ----------
    deployment : str
        The name of the deployment
    params_cache : Dict[str,Any], optional
        A populated dict with the key  of the parameter stored and its value
    session: Session, optional
        The boto3.Session to use to for SSM Parameter queries, default None

    Returns
    -------
    Optional[Dict[str, Any]]
        A dict of the deployment
    """
    return _fetch_helper(_deployed_deployment_manifest_key(deployment), params_cache, session=session)


def get_secret_secrets_manager(name: str, session: Optional[Session] = None) -> Dict[str, Any]:
    """
    get_secret_secrets_manager Fetches the data in the Secrets Manager with the key (name) given

    Parameters
    ----------
    name : str
        The name of the Secret in the Secrets manager
    session: Session, optional
        The boto3.Session to use to for SSM Parameter queries, default None

    Returns
    -------
    Dict[str, Any]
        The object in the Secrets Manager
    """
    return secrets.get_secrets_manager_value(name=name, session=session)


def get_secrets_version(
    secret_name: str,
    version_ref: Optional[str] = "AWSCURRENT",
    session: Optional[Session] = None,
) -> Optional[str]:
    versions = secrets.list_secret_version_ids(name=secret_name, session=session)

    if versions:
        for version in versions:
            version_id = version["VersionId"]
            version_stages = version["VersionStages"]
            if (version_id == version_ref) or (version_ref in version_stages):
                return str(version_id)
    return None


def get_ssm_parameter_version(
    ssm_parameter_name: str,
    session: Optional[Session] = None,
) -> Optional[int]:
    resp = ssm.describe_parameter(name=ssm_parameter_name, session=session)
    if resp is None:
        _logger.error("The SSM parameter %s could not be fetched", ssm_parameter_name)
        raise seedfarmer.errors.ModuleDeploymentError("The SSM parameter could not be fetched")
    else:
        return int(resp["Parameters"][0]["Version"]) if len(resp["Parameters"]) > 0 else None


def get_group_manifest(
    deployment: str, group: str, params_cache: Optional[Dict[str, Any]] = None, session: Optional[Session] = None
) -> Optional[Dict[str, Any]]:
    """
    get_group_manifest
        Get the group manifest stored for a deployed group

    Parameters
    ----------
    deployment : str
        The name of the deployment
    group : str
        The name of the group
    params_cache : Dict[str,Any], optional
        A populated dict with the key  of the parameter stored and its value
    session: Session, optional
        The boto3.Session to use to for SSM Parameter queries, default None

    Returns
    -------
    Optional[Dict[str, Any]]
        A dict of the group in a deployment
    """
    return _fetch_helper(_group_key(deployment, group), params_cache, session=session)


def does_md5_match(
    deployment: str,
    group: str,
    module: str,
    hash: str,
    type: ModuleConst,
    deployment_params_cache: Optional[Dict[str, Any]] = None,
    session: Optional[Session] = None,
) -> bool:
    """
    does_md5_match
        Check if a gneerated md5 hash matches what is currently deployed for a module

    Parameters
    ----------
    deployment : str
        The name of the deployment
    group : str
        The name of the group
    module : str
        The name of the module
    hash : str
        The generated md5 has you want to compare to what is currently dpeloyed
    type : ModuleConst
        An emumeration value of the md5 of the module you want
    params_cache : Dict[str,Any], optional
        A populated dict with the key  of the parameter stored and its value
    session: Session, optional
        The boto3.Session to use to for SSM Parameter queries, default None

    Returns
    -------
    bool
        Whether it does match or not
    """
    name = _md5_module_key(deployment, group, module, type)
    if not deployment_params_cache:
        p = ssm.get_parameter_if_exists(name=name, session=session)
    else:
        p = deployment_params_cache[name] if name in deployment_params_cache.keys() else None
    if not p:
        return False
    elif hash != p["hash"]:
        return False
    else:
        return True


def does_module_exist(deployment: str, group: str, module: str, session: Optional[Session] = None) -> bool:
    """
    does_module_exist
        Checks if a module of a group is deployed

    Parameters
    ----------
    deployment : str
        The name of the deployment
    group : str
        The name of the group
    module : str
        The name of the module
    session: Session, optional
        The boto3.Session to use to for SSM Parameter queries, default None

    Returns
    -------
    bool
        Whether the module is deployed
    """
    return ssm.does_parameter_exist(
        name=_md5_module_key(deployment, group, module, ModuleConst.BUNDLE), session=session
    )


def write_metadata(
    deployment: str, group: str, module: str, data: Dict[str, Any], session: Optional[Session] = None
) -> None:
    """
    write_metadata
        Persists the medadata of a deployed module

    Parameters
    ----------
    deployment : str
        The name of the deployment
    group : str
        The name of the group
    module : str
        The name of the module
    data : Dict[str, Any]
        The metadata of the module
    session: Session, optional
        The boto3.Session to use to for SSM Parameter queries, default None
    """
    ssm.put_parameter(name=_metadata_key(deployment, group, module), obj=data, session=session)


def write_group_manifest(deployment: str, group: str, data: Dict[str, Any], session: Optional[Session] = None) -> None:
    """
    write_group_manifest
        Persists the manifest of a deployed group

    Parameters
    ----------
    deployment : str
        The name of the deployment
    group : str
        The name of the group
    data : Dict[str, Any]
        The metadat of the module
    session: Session, optional
        The boto3.Session to use to for SSM Parameter queries, default None
    """
    ssm.put_parameter(name=_group_key(deployment, group), obj=data, session=session)


def write_module_manifest(
    deployment: str, group: str, module: str, data: Dict[str, Any], session: Optional[Session] = None
) -> None:
    """
    write_module_manifest
        Persists the manifest of a deployed module

    Parameters
    ----------
    deployment : str
        The name of the deployment
    group : str
        The name of the group
    module : str
        The name of the module
    data : Dict[str, Any]
        A dict of the data to be persisted
    session: Session, optional
        The boto3.Session to use to for SSM Parameter queries, default None
    """

    # Temp fix until a larger persistence store is vetted
    process_data = data
    current_size = sys.getsizeof(json.dumps(data))
    if current_size > 8191:
        _logger.info("The manifest for %s-%s is %s, too large for SSM, reducing", group, module, current_size)
        process_data = remove_nulls(data)
        _logger.info("The size is now %s", sys.getsizeof(json.dumps(process_data)))
    ssm.put_parameter(name=_manifest_key(deployment, group, module), obj=process_data, session=session)


def write_deployspec(
    deployment: str, group: str, module: str, data: Dict[str, Any], session: Optional[Session] = None
) -> None:
    """
    write_deployspec
        Persists the deployspec of a deployed module

    Parameters
    ----------
    deployment : str
        The name of the deployment
    group : str
        The name of the group
    module : str
        The name of the module
    data : Dict[str, Any]
        A dict of the data to be persisted
    session: Session, optional
        The boto3.Session to use to for SSM Parameter queries, default None
    """
    ssm.put_parameter(name=_deployspec_key(deployment, group, module), obj=data, session=session)


def write_module_md5(
    deployment: str, group: str, module: str, hash: str, type: ModuleConst, session: Optional[Session] = None
) -> None:
    """
    write_module_md5
        Persists the md5 of a module.
        There are three (3) relative md5 hashes of the module:
            - bundle / zip of currently deployed module
            - deployspec of currently deployed module
            - manifest of currently deployed module

    Parameters
    ----------
    deployment : str
        The name of the deployment
    group : str
        The name of the group
    module : str
        The name of the module
    hash : str
        The md5 has of the data
    type : ModuleConst
        An emumeration value of the md5 of the module this is
    session: Session, optional
        The boto3.Session to use to for SSM Parameter queries, default None
    """
    ssm.put_parameter(name=_md5_module_key(deployment, group, module, type), obj={"hash": hash}, session=session)


def write_deployment_manifest(deployment: str, data: Dict[str, Any], session: Optional[Session] = None) -> None:
    """
    write_deployment
        Persists the deployment manifest

    Parameters
    ----------
    deployment : str
        The name of the deployment
    data : Dict[str, Any]
        A dict of the deployment manifest
    session: Session, optional
        The boto3.Session to use to for SSM Parameter queries, default None
    """
    if _logger.isEnabledFor(logging.DEBUG):
        _logger.debug("Writing to %s values %s", _deployment_manifest_key(deployment), data)
    ssm.put_parameter(name=_deployment_manifest_key(deployment), obj=data, session=session)


def write_deployed_deployment_manifest(
    deployment: str, data: Dict[str, Any], session: Optional[Session] = None
) -> None:
    """
    write_deployed_deployment_manifest
        Persists the deployment manifest once all modules have been deployed

    Parameters
    ----------
    deployment : str
        The name of the deployment
    data : Dict[str, Any]
        A dict of the deployment manifest
    session: Session, optional
        The boto3.Session to use to for SSM Parameter queries, default None
    """
    key = _deployed_deployment_manifest_key(deployment)
    _logger.debug("Writing to %s value %s", key, data)

    ssm.put_parameter(name=key, obj=data, session=session)


def remove_module_info(deployment: str, group: str, module: str, session: Optional[Session] = None) -> None:
    """
    remove_module_info
        Delete all persisted data of a module in a group of a deployment

    Parameters
    ----------
    deployment : str
        The name of the deployment
    group : str
        The name of the group
    module : str
        The name of the module
    session: Session, optional
        The boto3.Session to use to for SSM Parameter queries, default None
    """
    ssm.delete_parameters(parameters=_all_module_keys(deployment, group, module), session=session)


def remove_group_info(deployment: str, group: str, session: Optional[Session] = None) -> None:
    """
    remove_group_info
        Delete all persisted data of a group of a deployment

    Parameters
    ----------
    deployment : str
        The name of the deployment
    group : str
        The name of the group
    session: Session, optional
        The boto3.Session to use to for SSM Parameter queries, default None
    """
    ssm.delete_parameters(parameters=(_all_group_keys(deployment, group)), session=session)


def remove_module_md5(
    deployment: str, group: str, module: str, type: ModuleConst, session: Optional[Session] = None
) -> None:
    """
    remove_module_md5
        Delete the md5 hash persisted of a module.
        There are three (3) relative md5 hashes of the module:
            - bundle / zip of currently deployed module
            - deployspec of currently deployed module
            - manifest of currently deployed module

    Parameters
    ----------
    deployment : str
        The name of the deployment
    group : str
        The name of the group
    module : str
        The name of the module
    type : ModuleConst
        An emumeration value of the md5 you want to delete
    session: Session, optional
        The boto3.Session to use to for SSM Parameter queries, default None
    """
    ssm.delete_parameters(parameters=[_md5_module_key(deployment, group, module, type)], session=session)


def remove_deployment_manifest(deployment: str, session: Optional[Session] = None) -> None:
    """
    remove_deployment_manifest
        Delete the deployment manifest persisted of a deployment

    Parameters
    ----------
    deployment : str
        The name of the deployment
    session: Session, optional
        The boto3.Session to use to for SSM Parameter queries, default None
    """
    ssm.delete_parameters(parameters=[_deployment_manifest_key(deployment)], session=session)


def remove_deployed_deployment_manifest(deployment: str, session: Optional[Session] = None) -> None:
    """
    remove_deployment_manifest
        Delete the deployment manifest persisted of a deployment

    Parameters
    ----------
    deployment : str
        The name of the deployment
    session: Session, optional
        The boto3.Session to use to for SSM Parameter queries, default None
    """
    ssm.delete_parameters(parameters=[_deployed_deployment_manifest_key(deployment)], session=session)


def _metadata_key(deployment: str, group: str, module: str) -> str:
    return f"/{config.PROJECT}/{deployment}/{group}/{module}/{ModuleConst.METADATA.value}"


def _md5_module_key(deployment: str, group: str, module: str, type: ModuleConst) -> str:
    return f"/{config.PROJECT}/{deployment}/{group}/{module}/{ModuleConst.MD5.value}/{type.value}"


def _md5_group_key(deployment: str, group: str, type: ModuleConst) -> str:
    return f"/{config.PROJECT}/{deployment}/{group}/{ModuleConst.MD5.value}/{type.value}"


def _deployspec_key(deployment: str, group: str, module: str) -> str:
    return f"/{config.PROJECT}/{deployment}/{group}/{module}/{ModuleConst.DEPLOYSPEC.value}"


def _manifest_key(deployment: str, group: str, module: str) -> str:
    return f"/{config.PROJECT}/{deployment}/{group}/{module}/{ModuleConst.MANIFEST.value}"


def _group_key(deployment: str, group: str) -> str:
    return f"/{config.PROJECT}/{deployment}/{group}/{ModuleConst.MANIFEST.value}"


def _deployment_manifest_key(deployment: str) -> str:
    return f"/{config.PROJECT}/{deployment}/{ModuleConst.MANIFEST.value}"


def _deployed_deployment_manifest_key(deployment: str) -> str:
    return f"/{config.PROJECT}/{deployment}/{ModuleConst.MANIFEST.value}/{ModuleConst.DEPLOYED.value}"


def _all_module_keys(deployment: str, group: str, module: str) -> List[str]:
    return [
        _metadata_key(deployment, group, module),
        _md5_module_key(deployment, group, module, ModuleConst.DEPLOYSPEC),
        _md5_module_key(deployment, group, module, ModuleConst.BUNDLE),
        _md5_module_key(deployment, group, module, ModuleConst.MANIFEST),
        _deployspec_key(deployment, group, module),
        _manifest_key(deployment, group, module),
    ]


def _all_group_keys(deployment: str, group: str) -> List[str]:
    return [_group_key(deployment, group), _md5_group_key(deployment, group, ModuleConst.MANIFEST)]


def _deployment_key(deployment: str) -> str:
    return f"/{config.PROJECT}/{deployment}/"


def _fetch_helper(
    name: str, params_cache: Optional[Dict[str, Any]] = None, session: Optional[Session] = None
) -> Optional[Dict[str, Any]]:
    if params_cache:
        return params_cache.get(name, None)
    else:
        return ssm.get_parameter_if_exists(name=name, session=session)


def get_module_stack_names(
    deployment_name: str, group_name: str, module_name: str, session: Optional[Session] = None
) -> Tuple[str, str]:
    resource_name = f"{config.PROJECT}-{deployment_name}-{group_name}-{module_name}"
    resource_hash = generate_hash(string=resource_name, length=4)
    # Max length of a Stack Name is 128 chars, -iam-policy is 11 chars, resource_hash plus "-" is 5 chars
    # If the resource_name and "-iam-policy" is too long, truncate and use a resource_hash for uniqueness
    module_stack_name = (
        f"{resource_name[: 128 - 11 - 5]}-{resource_hash}-iam-policy"
        if len(resource_name) > (128 - 11)
        else f"{resource_name}-iam-policy"
    )

    # Max length of a a Role Name is 64 chars, session_hash plus "-" is 9 chars, resource_hash plus "-" is 5 chars
    # If the resource_name and session_hash is too long, truncate and use a resource_hash for uniqueness
    session_hash = generate_session_hash(session=session)
    module_role_name = (
        f"{resource_name[: 64 - 9 - 5]}-{resource_hash}-{session_hash}"
        if len(resource_name) > (64 - 9)
        else f"{resource_name}-{session_hash}"
    )

    return module_stack_name, module_role_name


def get_modulestack_path(module_path: str) -> Any:
    p = os.path.join(config.OPS_ROOT, module_path, "modulestack.yaml")
    if not os.path.exists(p):
        _logger.debug("No modulestack.yaml found")
        return None
    return p


def get_deployspec_path(module_path: str) -> str:
    p = os.path.join(config.OPS_ROOT, module_path, "deployspec.yaml")
    if not os.path.exists(p):
        raise seedfarmer.errors.InvalidPathError(f"No deployspec.yaml file found in module directory: {p}")
    return os.path.join(config.OPS_ROOT, module_path, "deployspec.yaml")



#  Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License").
#    You may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

import json
import logging
import os
import subprocess
from pathlib import Path
from typing import Any, Dict, Optional, Tuple, cast

import yaml

from seedfarmer import config

# from seedfarmer.deployment.deploy_remote import _param as param
from seedfarmer.deployment.deploy_base import DeployModule
from seedfarmer.mgmt.module_info import get_deployspec_path
from seedfarmer.models._deploy_spec import DeploySpec

_logger: logging.Logger = logging.getLogger(__name__)


class ModuleMetadataSupport:
    use_project_prefix: Optional[bool] = None
    ops_root_path: Optional[str] = None
    project: Optional[str] = None

    def __init__(self) -> None:
        self.project = config.PROJECT
        self.ops_root_path = config.OPS_ROOT
        try:
            with open(get_deployspec_path("module")) as module_spec_file:
                deploy_spec = DeploySpec(**yaml.safe_load(module_spec_file))
            self.use_project_prefix = not deploy_spec.publish_generic_env_variables
        except Exception:
            _logger.warning("Cannot read the deployspec, using project name as prefix (a non-generic module)")
            self.use_project_prefix = True

    def get_ops_root_path(self) -> str:
        return str(self.ops_root_path)

    def metadata_file_name(self) -> str:
        return str(DeployModule.seedfarmer_param("MODULE_METADATA", None, self.use_project_prefix))

    def metadata_fullpath(self) -> str:
        return os.path.join(str(self.ops_root_path), "module", self.metadata_file_name())

    def project_param_name(self) -> str:
        return str(DeployModule.seedfarmer_param("PROJECT_NAME", None, self.use_project_prefix))

    def deployment_param_name(self) -> str:
        return str(DeployModule.seedfarmer_param("DEPLOYMENT_NAME", None, self.use_project_prefix))

    def module_param_name(self) -> str:
        return str(DeployModule.seedfarmer_param("MODULE_NAME", None, self.use_project_prefix))


def _read_metadata_file(mms: ModuleMetadataSupport) -> Dict[str, Any]:
    p = mms.metadata_fullpath()
    if Path(p).is_file():
        _logger.info("Reading metadata file at %s", p)
        try:
            with open(Path(mms.metadata_fullpath()), "r") as metadatafile:
                j = metadatafile.read()
            return cast(Dict[str, Any], json.loads(j))
        except json.decoder.JSONDecodeError:
            _logger.info("Cannot parse the file json")
            return {}
    else:
        _logger.info("Cannot find existing metadata file at %s, moving on", p)
        return {}


def _read_metadata_env_param(mms: ModuleMetadataSupport) -> Dict[str, Any]:
    p = mms.metadata_file_name()
    env_data = os.getenv(p)

    if env_data:
        try:
            return cast(Dict[str, Any], json.loads(env_data))
        except ValueError:
            try:
                return cast(Dict[str, Any], json.loads(env_data.replace("'", '"')))
            except Exception as e:
                _logger.info("Cannot parse env metadata %s after quote fix due to %s", p, e)
                return {}
        except Exception as e:
            _logger.info("Cannot parse env metadata %s due to %s", p, e)
            return {}
    else:
        _logger.info("Cannot find or empty metadata env param at %s, moving on", p)
        return {}


def _mod_dep_key(mms: ModuleMetadataSupport) -> str:
    return (
        f"{os.getenv(mms.project_param_name())}-"
        f"{os.getenv(mms.deployment_param_name())}-"
        f"{os.getenv(mms.module_param_name())}"
    )


def _read_json_file(mms: ModuleMetadataSupport, path: str) -> Tuple[str, Dict[str, Any]]:
    p = os.path.join(mms.get_ops_root_path(), "module", path)
    _logger.info("Reading extra file path located at %s", p)
    if os.path.isfile(p):
        with open(p, "r") as datafile:
            j = datafile.read()
        return p, dict(json.loads(j))
    else:
        return p, {}


def _write_metadata_file(mms: ModuleMetadataSupport, data: Dict[str, Any]) -> None:
    p = mms.metadata_fullpath()
    _logger.info("Writing metadata to %s", p)
    with open(p, "w") as outfile:
        outfile.write(json.dumps(data))


def _clean_jq(jq: str) -> str:
    sep = '"'
    a = [f"{sep}{val}{sep}" if val.find("-") != -1 else val for val in jq.split(".")]
    modified_str = ".".join(a)
    return f".{modified_str}" if not modified_str.startswith(".") else modified_str


def add_json_output(json_string: str) -> None:
    mms = ModuleMetadataSupport()
    json_new = json.loads(json_string)
    file_dict = _read_metadata_file(mms=mms)
    json_new = {**file_dict, **json_new} if file_dict else json_new
    _logger.debug(f"Current Dict {json.dumps(json_new, indent=4)}")
    env_dict = _read_metadata_env_param(mms=mms)
    json_new = {**env_dict, **json_new} if env_dict else json_new
    _logger.debug(f"Current Dict {json.dumps(json_new, indent=4)}")
    _write_metadata_file(mms=mms, data=json_new)


def add_kv_output(key: str, value: str) -> None:
    mms = ModuleMetadataSupport()
    data = {}
    data[key] = value
    file_dict = _read_metadata_file(mms=mms)
    data = {**file_dict, **data} if file_dict else data
    _logger.debug(f"Current Dict {json.dumps(data, indent=4)}")
    env_dict = _read_metadata_env_param(mms=mms)
    data = {**env_dict, **data} if env_dict else data
    _logger.debug(f"Current Dict {json.dumps(data, indent=4)}")
    _write_metadata_file(mms=mms, data=data)


def convert_cdkexports(
    json_file: str,
    jq_path: Optional[str] = None,
) -> None:
    mms = ModuleMetadataSupport()
    cdk_output_path, cdk_output = _read_json_file(mms, json_file)
    data = {}
    if cdk_output:
        if not jq_path:
            out_key = _mod_dep_key(mms)
            _logger.info("Pulling %s from the %s file", out_key, json_file)
            data = cdk_output[out_key]["metadata"]
        else:
            clean_jq_path = _clean_jq(jq_path)
            _logger.info("Pulling with jq path '%s' from %s file", clean_jq_path, json_file)
            with open(cdk_output_path, "r") as infile, open("tmp-metadata", "w") as outfile:
                subprocess.run(["jq", clean_jq_path], stdin=infile, stdout=outfile, shell=False)
            data = json.loads(open("tmp-metadata", "r").read())

    existing_metadata = _read_metadata_file(mms)

    try:
        _write_metadata_file(mms=mms, data={**existing_metadata, **json.loads(data)}) if data else None  # type: ignore[arg-type]
    except json.decoder.JSONDecodeError:
        _logger.info("The CDK Export is not a string that can be converted to a JSON, ignoring this additional data")
        _logger.info("Offending metadata -- %s", data)
        _write_metadata_file(mms=mms, data=existing_metadata) if existing_metadata else None


def get_dep_mod_name() -> str:
    return _mod_dep_key(ModuleMetadataSupport())


def get_parameter_value(parameter_suffix: str) -> Optional[str]:
    mms = ModuleMetadataSupport()
    key = DeployModule.seedfarmer_param(parameter_suffix, None, mms.use_project_prefix)
    try:
        _logger.info("Getting the Env Parameter tied to %s", key)
        return os.getenv(key)
    except Exception:
        _logger.warning("Error looking for %s, returning None", key)
        return None



#  Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License").
#    You may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

import os
import sys

import yaml

from seedfarmer import CLI_ROOT, DEFAULT_PROJECT_POLICY_PATH


def get_default_project_policy(policy_prefix: str = "/") -> None:
    with open(os.path.join(CLI_ROOT, DEFAULT_PROJECT_POLICY_PATH), "rb") as f:
        policy = yaml.safe_load(f)

        if policy_prefix:
            policy["Resources"]["ProjectPolicy"]["Properties"]["Path"] = policy_prefix

        sys.stdout.write(yaml.dump(policy))
